from vyper import ast as vy_ast
from vyper.compiler.settings import get_global_settings
from vyper.exceptions import (
    ArrayIndexException,
    CompilerPanic,
    FeatureException,
    InstantiationException,
    InvalidType,
    StructureException,
    UndeclaredDefinition,
    UnknownType,
)
from vyper.semantics.analysis.levenshtein_utils import get_levenshtein_error_suggestions
from vyper.semantics.data_locations import DataLocation
from vyper.semantics.namespace import get_namespace
from vyper.semantics.types.base import TYPE_T, VyperType

# TODO maybe this should be merged with .types/base.py


def type_from_abi(abi_type: dict) -> VyperType:
    """
    Return a type object from an ABI type definition.

    Arguments
    ---------
    abi_type : dict
       A type definition taken from the `input` or `output` field of an ABI.

    Returns
    -------
    BaseTypeDefinition
        Type definition object.
    """
    type_string = abi_type["type"]
    if type_string == "int168" and abi_type.get("internalType") == "decimal":
        type_string = "decimal"
    if type_string in ("string", "bytes"):
        type_string = type_string.capitalize()

    namespace = get_namespace()

    if "[" in type_string:
        # handle dynarrays, static arrays
        value_type_string, length_str = type_string.rsplit("[", maxsplit=1)
        try:
            length = int(length_str.rstrip("]"))
        except ValueError:
            raise UnknownType(f"ABI type has an invalid length: {type_string}") from None
        try:
            value_type = type_from_abi({"type": value_type_string})
        except UnknownType:
            raise UnknownType(f"ABI contains unknown type: {type_string}") from None
        try:
            sarray_t = namespace["$SArrayT"]
            return sarray_t(value_type, length)
        except InvalidType:
            raise UnknownType(f"ABI contains unknown type: {type_string}") from None

    else:
        try:
            t = namespace[type_string]
            if type_string in ("Bytes", "String"):
                # special handling for bytes, string, since
                # the type ctor is in the namespace instead of a concrete type.
                return t()
            return t
        except KeyError:
            raise UnknownType(f"ABI contains unknown type: {type_string}") from None


def type_from_annotation(
    node: vy_ast.VyperNode, location: DataLocation = DataLocation.UNSET
) -> VyperType:
    """
    Return a type object for the given AST node after validating its location.

    Arguments
    ---------
    node: VyperNode
        Vyper ast node from the `annotation` member of a `VariableDecl` or `AnnAssign` node.

    Returns
    -------
    VyperType
        Type definition object.
    """
    typ = _type_from_annotation(node)

    if location in typ._invalid_locations:
        location_str = "" if location is DataLocation.UNSET else f"in {location.name.lower()}"
        raise InstantiationException(f"{typ} is not instantiable {location_str}", node)

    # TODO: cursed import cycle!
    from vyper.semantics.types.primitives import DecimalT

    if isinstance(typ, DecimalT):
        # is there a better place to put this check?
        settings = get_global_settings()
        if settings and not settings.get_enable_decimals():
            raise FeatureException("decimals are not allowed unless `--enable-decimals` is set")

    return typ


def _type_from_annotation(node: vy_ast.VyperNode) -> VyperType:
    namespace = get_namespace()

    if isinstance(node, vy_ast.Tuple):
        tuple_t = namespace["$TupleT"]
        return tuple_t.from_annotation(node)

    if isinstance(node, vy_ast.Subscript):
        # ex. HashMap, DynArray, Bytes, static arrays
        if node.value.get("id") in ("HashMap", "Bytes", "String", "DynArray"):
            assert isinstance(node.value, vy_ast.Name)  # mypy hint
            type_ctor = namespace[node.value.id]
        else:
            # like, address[5] or int256[5][5]
            type_ctor = namespace["$SArrayT"]

        return type_ctor.from_annotation(node)

    # prepare a common error message
    err_msg = f"'{node.node_source_code}' is not a type!"

    if isinstance(node, vy_ast.Attribute):
        # ex. SomeModule.SomeStruct

        if isinstance(node.value, vy_ast.Attribute):
            module_or_interface = _type_from_annotation(node.value)
        elif isinstance(node.value, vy_ast.Name):
            try:
                module_or_interface = namespace[node.value.id]  # type: ignore
            except UndeclaredDefinition:
                raise InvalidType(err_msg, node) from None
        else:
            raise InvalidType(err_msg, node)

        if hasattr(module_or_interface, "module_t"):  # i.e., it's a ModuleInfo
            module_or_interface = module_or_interface.module_t

        if not isinstance(module_or_interface, VyperType):
            raise InvalidType(err_msg, node)

        if not module_or_interface._attribute_in_annotation:
            raise InvalidType(err_msg, node)

        type_t = module_or_interface.get_type_member(node.attr, node)  # type: ignore
        assert isinstance(type_t, TYPE_T)  # sanity check
        return type_t.typedef

    if not isinstance(node, vy_ast.Name):
        # maybe handle this somewhere upstream in ast validation
        raise InvalidType(err_msg, node)

    if node.id not in namespace:  # type: ignore
        hint = get_levenshtein_error_suggestions(node.node_source_code, namespace, 0.3)
        raise UnknownType(
            f"No builtin or user-defined type named '{node.node_source_code}'.", node, hint=hint
        ) from None

    typ_ = namespace[node.id]
    if hasattr(typ_, "from_annotation"):
        # cases where the object in the namespace is an uninstantiated
        # type object, ex. Bytestring or DynArray (with no length provided).
        # call from_annotation to produce a better error message.
        typ_.from_annotation(node)

    if hasattr(typ_, "module_t"):  # it's a ModuleInfo
        typ_ = typ_.module_t

    if not isinstance(typ_, VyperType):
        raise CompilerPanic(f"Not a type: {typ_}", node)

    return typ_


def get_index_value(node: vy_ast.VyperNode) -> int:
    """
    Return the literal value for a `Subscript` index.

    Arguments
    ---------
    node: vy_ast.VyperNode
        Vyper ast node from the `slice` member of a Subscript node.

    Returns
    -------
    int
        Literal integer value.
        In the future, will return `None` if the subscript is an Ellipsis
    """
    # this is imported to improve error messages
    # TODO: revisit this!
    from vyper.semantics.analysis.utils import get_possible_types_from_node

    node = node.reduced()

    if not isinstance(node, vy_ast.Int):
        # even though the subscript is an invalid type, first check if it's a valid _something_
        # this gives a more accurate error in case of e.g. a typo in a constant variable name
        try:
            get_possible_types_from_node(node)
        except StructureException:
            # StructureException is a very broad error, better to raise InvalidType in this case
            pass
        raise InvalidType("Subscript must be a literal integer", node)

    if node.value <= 0:
        raise ArrayIndexException("Subscript must be greater than 0", node)

    return node.value



from typing import Iterator, Optional

from vyper.codegen.ir_node import IRnode
from vyper.utils import OrderedSet
from vyper.venom.basicblock import CFG_ALTERING_INSTRUCTIONS, IRBasicBlock, IRLabel, IRVariable


class IRFunction:
    """
    Function that contains basic blocks.
    """

    name: IRLabel  # symbol name
    ctx: "IRContext"  # type: ignore # noqa: F821
    args: list
    last_label: int
    last_variable: int
    _basic_block_dict: dict[str, IRBasicBlock]

    # Used during code generation
    _ast_source_stack: list[IRnode]
    _error_msg_stack: list[str]

    def __init__(self, name: IRLabel, ctx: "IRContext" = None) -> None:  # type: ignore # noqa: F821
        self.ctx = ctx
        self.name = name
        self.args = []
        self._basic_block_dict = {}

        self.last_variable = 0

        self._ast_source_stack = []
        self._error_msg_stack = []

        self.append_basic_block(IRBasicBlock(name, self))

    @property
    def entry(self) -> IRBasicBlock:
        return next(self.get_basic_blocks())

    def append_basic_block(self, bb: IRBasicBlock):
        """
        Append basic block to function.
        """
        assert isinstance(bb, IRBasicBlock), bb
        assert bb.label.name not in self._basic_block_dict
        self._basic_block_dict[bb.label.name] = bb

    def remove_basic_block(self, bb: IRBasicBlock):
        assert isinstance(bb, IRBasicBlock), bb
        del self._basic_block_dict[bb.label.name]

    def get_basic_block(self, label: Optional[str] = None) -> IRBasicBlock:
        """
        Get basic block by label.
        If label is None, return the last basic block.
        """
        if label is None:
            return next(reversed(self._basic_block_dict.values()))

        return self._basic_block_dict[label]

    def clear_basic_blocks(self):
        self._basic_block_dict.clear()

    def get_basic_blocks(self) -> Iterator[IRBasicBlock]:
        """
        Get an iterator over this function's basic blocks
        """
        return iter(self._basic_block_dict.values())

    @property
    def num_basic_blocks(self) -> int:
        return len(self._basic_block_dict)

    def get_terminal_basicblocks(self) -> Iterator[IRBasicBlock]:
        """
        Get basic blocks that are terminal.
        """
        for bb in self.get_basic_blocks():
            if bb.is_terminal:
                yield bb

    def get_next_variable(self) -> IRVariable:
        self.last_variable += 1
        return IRVariable(f"%{self.last_variable}")

    def get_last_variable(self) -> str:
        return f"%{self.last_variable}"

    def remove_unreachable_blocks(self) -> int:
        self._compute_reachability()

        removed = []

        # Remove unreachable basic blocks
        for bb in self.get_basic_blocks():
            if not bb.is_reachable:
                removed.append(bb)

        for bb in removed:
            self.remove_basic_block(bb)

        # Remove phi instructions that reference removed basic blocks
        for bb in removed:
            for out_bb in bb.cfg_out:
                out_bb.remove_cfg_in(bb)
                for inst in out_bb.instructions:
                    if inst.opcode != "phi":
                        continue
                    in_labels = inst.get_label_operands()
                    if bb.label in in_labels:
                        inst.remove_phi_operand(bb.label)
                    op_len = len(inst.operands)
                    if op_len == 2:
                        inst.opcode = "store"
                        inst.operands = [inst.operands[1]]
                    elif op_len == 0:
                        out_bb.remove_instruction(inst)

        return len(removed)

    def _compute_reachability(self) -> None:
        """
        Compute reachability of basic blocks.
        """
        for bb in self.get_basic_blocks():
            bb.reachable = OrderedSet()
            bb.is_reachable = False

        self._compute_reachability_from(self.entry)

    def _compute_reachability_from(self, bb: IRBasicBlock) -> None:
        """
        Compute reachability of basic blocks from bb.
        """
        if bb.is_reachable:
            return
        bb.is_reachable = True
        for inst in bb.instructions:
            if inst.opcode in CFG_ALTERING_INSTRUCTIONS:
                for op in inst.get_label_operands():
                    out_bb = self.get_basic_block(op.value)
                    bb.reachable.add(out_bb)
                    self._compute_reachability_from(out_bb)

    @property
    def normalized(self) -> bool:
        """
        Check if function is normalized. A function is normalized if in the
        CFG, no basic block simultaneously has multiple inputs and outputs.
        That is, a basic block can be jumped to *from* multiple blocks, or it
        can jump *to* multiple blocks, but it cannot simultaneously do both.
        Having a normalized CFG makes calculation of stack layout easier when
        emitting assembly.
        """
        for bb in self.get_basic_blocks():
            # Ignore if there are no multiple predecessors
            if len(bb.cfg_in) <= 1:
                continue

            # Check if there is a branching jump at the end
            # of one of the predecessors
            for in_bb in bb.cfg_in:
                if len(in_bb.cfg_out) > 1:
                    return False

        # The function is normalized
        return True

    def push_source(self, ir):
        if isinstance(ir, IRnode):
            self._ast_source_stack.append(ir.ast_source)
            self._error_msg_stack.append(ir.error_msg)

    def pop_source(self):
        assert len(self._ast_source_stack) > 0, "Empty source stack"
        self._ast_source_stack.pop()
        assert len(self._error_msg_stack) > 0, "Empty error stack"
        self._error_msg_stack.pop()

    @property
    def ast_source(self) -> Optional[IRnode]:
        return self._ast_source_stack[-1] if len(self._ast_source_stack) > 0 else None

    @property
    def error_msg(self) -> Optional[str]:
        return self._error_msg_stack[-1] if len(self._error_msg_stack) > 0 else None

    def chain_basic_blocks(self) -> None:
        """
        Chain basic blocks together. If a basic block is not terminated, jump to the next one.
        Otherwise, append a stop instruction. This is necessary for the IR to be valid, and is
        done after the IR is generated.
        """
        bbs = list(self.get_basic_blocks())
        for i, bb in enumerate(bbs):
            if not bb.is_terminated:
                if i < len(bbs) - 1:
                    # TODO: revisit this. When contructor calls internal functions they
                    # are linked to the last ctor block. Should separate them before this
                    # so we don't have to handle this here
                    if bbs[i + 1].label.value.startswith("internal"):
                        bb.append_instruction("stop")
                    else:
                        bb.append_instruction("jmp", bbs[i + 1].label)
                else:
                    bb.append_instruction("exit")

    def copy(self):
        new = IRFunction(self.name)
        new._basic_block_dict = self._basic_block_dict.copy()
        new.last_label = self.last_label
        new.last_variable = self.last_variable
        return new

    def as_graph(self, only_subgraph=False) -> str:
        """
        Return the function as a graphviz dot string. If only_subgraph is True, only return the
        subgraph, not the full digraph -for embedding in a larger graph-
        """
        import html

        def _make_label(bb):
            ret = '<<table border="1" cellborder="0" cellspacing="0">'
            ret += f'<tr><td align="left"><b>{html.escape(str(bb.label))}</b></td></tr>\n'
            for inst in bb.instructions:
                ret += f'<tr ><td align="left">{html.escape(str(inst))}</td></tr>\n'
            ret += "</table>>"

            return ret
            # return f"{bb.label.value}:\n" + "\n".join([f"    {inst}" for inst in bb.instructions])

        ret = []

        if not only_subgraph:
            ret.append("digraph G {{")
        ret.append(f'subgraph "{self.name}" {{')

        for bb in self.get_basic_blocks():
            for out_bb in bb.cfg_out:
                ret.append(f'    "{bb.label.value}" -> "{out_bb.label.value}"')

        for bb in self.get_basic_blocks():
            ret.append(f'    "{bb.label.value}" [shape=plaintext, ')
            ret.append(f'label={_make_label(bb)}, fontname="Courier" fontsize="8"]')

        ret.append("}\n")
        if not only_subgraph:
            ret.append("}\n")

        return "\n".join(ret)

    def __repr__(self) -> str:
        str = f"IRFunction: {self.name}\n"
        for bb in self.get_basic_blocks():
            str += f"{bb}\n"
        return str.strip()



import contextlib
import copy
import re
from enum import Enum, auto
from functools import cached_property
from typing import Any, List, Optional, Union

import vyper.ast as vy_ast
from vyper.compiler.settings import VYPER_COLOR_OUTPUT
from vyper.evm.address_space import AddrSpace
from vyper.evm.opcodes import get_ir_opcodes
from vyper.exceptions import CodegenPanic, CompilerPanic
from vyper.semantics.types import VyperType
from vyper.utils import VALID_IR_MACROS, ceil32

# Set default string representation for ints in IR output.
AS_HEX_DEFAULT = False

if VYPER_COLOR_OUTPUT:
    OKBLUE = "\033[94m"
    OKMAGENTA = "\033[35m"
    OKLIGHTMAGENTA = "\033[95m"
    OKLIGHTBLUE = "\033[94m"
    ENDC = "\033[0m"
else:
    OKBLUE = ""
    OKMAGENTA = ""
    OKLIGHTMAGENTA = ""
    OKLIGHTBLUE = ""
    ENDC = ""


class NullAttractor(int):
    def __add__(self, other: int) -> "NullAttractor":
        return NullAttractor()

    def __repr__(self) -> str:
        return "None"

    __radd__ = __add__
    __mul__ = __add__


class Encoding(Enum):
    # vyper encoding, default for memory variables
    VYPER = auto()
    # abi encoded, default for args/return values from external funcs
    ABI = auto()
    # future: packed


# shortcut for chaining multiple cache_when_complex calls
# CMC 2023-08-10 remove this _as soon as_ we have
# real variables in IR (that we can declare without explicit scoping -
# needs liveness analysis).
@contextlib.contextmanager
def scope_multi(ir_nodes, names):
    assert len(ir_nodes) == len(names)

    builders = []
    scoped_ir_nodes = []

    class _MultiBuilder:
        def resolve(self, body):
            # sanity check that it's initialized properly
            assert len(builders) == len(ir_nodes)
            ret = body
            for b in reversed(builders):
                ret = b.resolve(ret)
            return ret

    mb = _MultiBuilder()

    with contextlib.ExitStack() as stack:
        for arg, name in zip(ir_nodes, names):
            b, ir_node = stack.enter_context(arg.cache_when_complex(name))

            builders.append(b)
            scoped_ir_nodes.append(ir_node)

        yield mb, scoped_ir_nodes


# this creates a magical block which maps to IR `with`
class _WithBuilder:
    def __init__(self, ir_node, name, should_inline=False):
        if should_inline and ir_node._optimized.is_complex_ir:  # pragma: nocover
            # this can only mean trouble
            raise CompilerPanic("trying to inline a complex IR node")

        self.ir_node = ir_node

        # whether or not to inline the ir_node
        self.should_inline = should_inline

        # a named IR variable which represents the
        # output of `ir_node`
        self.ir_var = IRnode.from_list(
            name, typ=ir_node.typ, location=ir_node.location, encoding=ir_node.encoding
        )

    def __enter__(self):
        if self.should_inline:
            # return the value instead of the named variable
            # so it can be inlined
            return self, self.ir_node
        else:
            # return the named variable
            return self, self.ir_var

    def __exit__(self, *args):
        pass

    # MUST be called at the end of building the expression
    # in order to make sure the expression gets wrapped correctly
    def resolve(self, body):
        if self.should_inline:
            # uses of the variable have already been inlined
            return body

        ret = ["with", self.ir_var, self.ir_node, body]
        if isinstance(body, IRnode):
            return IRnode.from_list(
                ret, typ=body.typ, location=body.location, encoding=body.encoding
            )
        else:
            return ret


# Data structure for IR parse tree
class IRnode:
    repr_show_gas = False
    _gas: int
    valency: int
    args: List["IRnode"]
    value: Union[str, int]
    is_self_call: bool
    passthrough_metadata: dict[str, Any]
    func_ir: Any
    common_ir: Any

    def __init__(
        self,
        value: Union[str, int],
        args: List["IRnode"] = None,
        typ: VyperType = None,
        location: Optional[AddrSpace] = None,
        ast_source: Optional[vy_ast.VyperNode] = None,
        annotation: Optional[str] = None,
        error_msg: Optional[str] = None,
        mutable: bool = True,
        add_gas_estimate: int = 0,
        encoding: Encoding = Encoding.VYPER,
        is_self_call: bool = False,
        passthrough_metadata: dict[str, Any] = None,
    ):
        if args is None:
            args = []

        self.value = value
        self.args = args
        # TODO remove this sanity check once mypy is more thorough
        assert isinstance(typ, VyperType) or typ is None, repr(typ)
        self.typ = typ
        self.location = location
        self.ast_source = ast_source
        self.error_msg = error_msg
        self.annotation = annotation
        self.mutable = mutable
        self.add_gas_estimate = add_gas_estimate
        self.encoding = encoding
        self.as_hex = AS_HEX_DEFAULT
        self.is_self_call = is_self_call
        self.passthrough_metadata = passthrough_metadata or {}
        self.func_ir = None
        self.common_ir = None

        assert self.value is not None, "None is not allowed as IRnode value"

        # Determine this node's valency (1 if it pushes a value on the stack,
        # 0 otherwise) and checks to make sure the number and valencies of
        # children are correct. Also, find an upper bound on gas consumption
        # Numbers
        if isinstance(self.value, int):
            assert len(self.args) == 0, "int can't have arguments"

            # integers must be in the range (MIN_INT256, MAX_UINT256)
            assert -(2**255) <= self.value < 2**256, "out of range"

            self.valency = 1
            self._gas = 5
        elif isinstance(self.value, bytes):
            # a literal bytes value, probably inside a "data" node.
            assert len(self.args) == 0, "bytes can't have arguments"

            self.valency = 0
            self._gas = 0

        elif isinstance(self.value, str):
            # Opcodes and pseudo-opcodes (e.g. clamp)
            if self.value.upper() in get_ir_opcodes():
                _, ins, outs, gas = get_ir_opcodes()[self.value.upper()]
                self.valency = outs
                assert (
                    len(self.args) == ins
                ), f"Number of arguments mismatched: {self.value} {self.args}"
                # We add 2 per stack height at push time and take it back
                # at pop time; this makes `break` easier to handle
                self._gas = gas + 2 * (outs - ins)
                for arg in self.args:
                    # pop and pass are used to push/pop values on the stack to be
                    # consumed for internal functions, therefore we whitelist this as a zero valency
                    # allowed argument.
                    zero_valency_whitelist = {"pass", "pop"}
                    assert (
                        arg.valency == 1 or arg.value in zero_valency_whitelist
                    ), f"invalid argument to `{self.value}`: {arg}"

                    self._gas += arg.gas
                # Dynamic gas cost: 8 gas for each byte of logging data
                if self.value.upper()[0:3] == "LOG" and isinstance(self.args[1].value, int):
                    self._gas += self.args[1].value * 8
                # Dynamic gas cost: non-zero-valued call
                if self.value.upper() == "CALL" and self.args[2].value != 0:
                    self._gas += 34000
                # Dynamic gas cost: filling sstore (ie. not clearing)
                elif self.value.upper() == "SSTORE" and self.args[1].value != 0:
                    self._gas += 15000
                # Dynamic gas cost: calldatacopy
                elif self.value.upper() in ("CALLDATACOPY", "CODECOPY", "EXTCODECOPY"):
                    size = 34000
                    size_arg_index = 3 if self.value.upper() == "EXTCODECOPY" else 2
                    size_arg = self.args[size_arg_index]
                    if isinstance(size_arg.value, int):
                        size = size_arg.value
                    self._gas += ceil32(size) // 32 * 3
                # Gas limits in call
                if self.value.upper() == "CALL" and isinstance(self.args[0].value, int):
                    self._gas += self.args[0].value
            # If statements
            elif self.value == "if":
                if len(self.args) == 3:
                    self._gas = self.args[0].gas + max(self.args[1].gas, self.args[2].gas) + 3
                if len(self.args) == 2:
                    self._gas = self.args[0].gas + self.args[1].gas + 17
                assert (
                    self.args[0].valency > 0
                ), f"zerovalent argument as a test to an if statement: {self.args[0]}"
                assert len(self.args) in (2, 3), "if statement can only have 2 or 3 arguments"
                self.valency = self.args[1].valency
            # With statements: with <var> <initial> <statement>
            elif self.value == "with":
                assert len(self.args) == 3, self
                assert len(self.args[0].args) == 0 and isinstance(
                    self.args[0].value, str
                ), f"first argument to with statement must be a variable name: {self.args[0]}"
                assert (
                    self.args[1].valency == 1 or self.args[1].value == "pass"
                ), f"zerovalent argument to with statement: {self.args[1]}"
                self.valency = self.args[2].valency
                self._gas = sum([arg.gas for arg in self.args]) + 5
            # Repeat statements: repeat <index_name> <startval> <rounds> <rounds_bound> <body>
            elif self.value == "repeat":
                assert (
                    len(self.args) == 5
                ), "repeat(index_name, startval, rounds, rounds_bound, body)"

                counter_ptr = self.args[0]
                start = self.args[1]
                repeat_count = self.args[2]
                repeat_bound = self.args[3]
                body = self.args[4]

                assert (
                    isinstance(repeat_bound.value, int) and repeat_bound.value > 0
                ), f"repeat bound must be a compile-time positive integer: {self.args[2]}"
                assert repeat_count.valency == 1, repeat_count
                assert counter_ptr.valency == 1, counter_ptr
                assert start.valency == 1, start

                self.valency = 0

                self._gas = counter_ptr.gas + start.gas
                self._gas += 3  # gas for repeat_bound
                int_bound = int(repeat_bound.value)
                self._gas += int_bound * (body.gas + 50) + 30

                if repeat_count != repeat_bound:
                    # gas for assert(repeat_count <= repeat_bound)
                    self._gas += 18

            # Seq statements: seq <statement> <statement> ...
            elif self.value == "seq":
                self.valency = self.args[-1].valency if self.args else 0
                self._gas = sum([arg.gas for arg in self.args]) + 30

            # GOTO is a jump with args
            # e.g. (goto my_label x y z) will push x y and z onto the stack,
            # then JUMP to my_label.
            elif self.value in ("goto", "exit_to"):
                for arg in self.args:
                    assert (
                        arg.valency == 1 or arg.value == "pass"
                    ), f"zerovalent argument to goto {arg}"

                self.valency = 0
                self._gas = sum([arg.gas for arg in self.args])
            elif self.value == "label":
                assert (
                    self.args[1].value == "var_list"
                ), f"2nd argument to label must be var_list, {self}"
                assert len(args) == 3, f"label should have 3 args but has {len(args)}, {self}"
                self.valency = 0
                self._gas = 1 + sum(t.gas for t in self.args)
            elif self.value == "unique_symbol":
                # a label which enforces uniqueness, and does not appear
                # in generated bytecode. this is useful for generating
                # internal assertions that a particular IR fragment only
                # occurs a single time in a program. note that unique_symbol
                # must be distinct from all `unique_symbol`s AS WELL AS all
                # `label`s, otherwise IR-to-assembly will raise an exception.
                self.valency = 0
                self._gas = 0

            # var_list names a variable number stack variables
            elif self.value == "var_list":
                for arg in self.args:
                    if not isinstance(arg.value, str) or len(arg.args) > 0:  # pragma: nocover
                        raise CodegenPanic(f"var_list only takes strings: {self.args}")
                self.valency = 0
                self._gas = 0

            # Multi statements: multi <expr> <expr> ...
            elif self.value == "multi":
                for arg in self.args:
                    assert (
                        arg.valency > 0
                    ), f"Multi expects all children to not be zerovalent: {arg}"
                self.valency = sum([arg.valency for arg in self.args])
                self._gas = sum([arg.gas for arg in self.args])
            elif self.value == "deploy":
                self.valency = 0
                assert len(self.args) == 3, f"`deploy` should have three args {self}"
                self._gas = NullAttractor()  # unknown
            # Stack variables
            else:
                self.valency = 1
                self._gas = 3
        else:  # pragma: nocover
            raise CompilerPanic(f"Invalid value for IR AST node: {self.value}")
        assert isinstance(self.args, list)

    # deepcopy is a perf hotspot; it pays to optimize it a little
    def __deepcopy__(self, memo):
        cls = self.__class__
        ret = cls.__new__(cls)
        ret.__dict__ = self.__dict__.copy()
        ret.args = [copy.deepcopy(arg) for arg in ret.args]
        return ret

    # TODO would be nice to rename to `gas_estimate` or `gas_bound`
    @property
    def gas(self):
        return self._gas + self.add_gas_estimate

    # the IR should be cached and/or evaluated exactly once
    @property
    def is_complex_ir(self):
        # list of items not to cache. note can add other env variables
        # which do not change, e.g. calldatasize, coinbase, etc.
        # reads (from memory or storage) should not be cached because
        # they can have or be affected by side effects.
        do_not_cache = {"~empty", "calldatasize", "callvalue"}

        return (
            isinstance(self.value, str)
            and (self.value.lower() in VALID_IR_MACROS or self.value.upper() in get_ir_opcodes())
            and self.value.lower() not in do_not_cache
        )

    # set an error message and push down into all children.
    # useful for overriding an error message generated by a helper
    # function with a more specific error message.
    def set_error_msg(self, error_msg: str) -> None:
        self.error_msg = error_msg
        for arg in self.args:
            arg.set_error_msg(error_msg)

    # get the unique symbols contained in this node, which provides
    # sanity check invariants for the optimizer.
    # cache because it's a perf hotspot. note that this (and other cached
    # properties!) can get borked if `self.args` are mutated in such a way
    # which changes the child `.unique_symbols`. in the future it would
    # be good to tighten down the hatches so it is harder to modify
    # IRnode member variables.
    @cached_property
    def unique_symbols(self):
        ret = set()
        if self.value == "unique_symbol":
            ret.add(self.args[0].value)

        children = self.args
        if self.value == "deploy":
            children = [self.args[0], self.args[2]]
        for arg in children:
            s = arg.unique_symbols
            non_uniques = ret.intersection(s)
            assert len(non_uniques) == 0, f"non-unique symbols {non_uniques}"
            ret |= s
        return ret

    @property
    def is_literal(self) -> bool:
        return isinstance(self.value, int) or self.value == "multi"

    def int_value(self) -> int:
        assert isinstance(self.value, int)
        return self.value

    @property
    def is_pointer(self) -> bool:
        # not used yet but should help refactor/clarify downstream code
        # eventually
        return self.location is not None

    @property  # probably could be cached_property but be paranoid
    def _optimized(self):
        # TODO figure out how to fix this circular import
        from vyper.ir.optimizer import optimize

        return optimize(self)

    # This function is slightly confusing but abstracts a common pattern:
    # when an IR value needs to be computed once and then cached as an
    # IR value (if it is expensive, or more importantly if its computation
    # includes side-effects), cache it as an IR variable named with the
    # `name` param, and execute the `body` with the cached value. Otherwise,
    # run the `body` without caching the IR variable.
    # Note that this may be an unneeded abstraction in the presence of an
    # arbitrarily powerful optimization framework (which can detect unneeded
    # caches) but for now still necessary - CMC 2021-12-11.
    # usage:
    # ```
    # with ir_node.cache_when_complex("foo") as builder, foo:
    #   ret = some_function(foo)
    #   return builder.resolve(ret)
    # ```
    def cache_when_complex(self, name):
        # for caching purposes, see if the ir_node will be optimized
        # because a non-literal expr could turn into a literal,
        # (e.g. `(add 1 2)`)
        # TODO this could really be moved into optimizer.py
        should_inline = not self._optimized.is_complex_ir

        return _WithBuilder(self, name, should_inline)

    @cached_property
    def referenced_variables(self):
        ret = set()
        for arg in self.args:
            ret |= arg.referenced_variables

        ret |= getattr(self, "_referenced_variables", set())

        return ret

    @cached_property
    def contains_self_call(self):
        return getattr(self, "is_self_call", False) or any(x.contains_self_call for x in self.args)

    # TODO this seems like a not useful and also confusing function
    # check if dead code and remove - CMC 2021-12-13
    def to_list(self):
        return [self.value] + [a.to_list() for a in self.args]

    def __eq__(self, other):
        return (
            self.value == other.value
            and self.args == other.args
            and self.typ == other.typ
            and self.location == other.location
            and self.mutable == other.mutable
            and self.add_gas_estimate == other.add_gas_estimate
        )

    @property
    def repr_value(self):
        if isinstance(self.value, int) and self.as_hex:
            return hex(self.value)
        if not isinstance(self.value, str):
            return str(self.value)
        return self.value

    @staticmethod
    def _colorise_keywords(val):
        if val.lower() in VALID_IR_MACROS:  # highlight macro
            return OKLIGHTMAGENTA + val + ENDC
        elif val.upper() in get_ir_opcodes().keys():
            return OKMAGENTA + val + ENDC
        return val

    def repr(self) -> str:
        if not len(self.args):
            if self.annotation:
                return f"{self.repr_value} " + OKLIGHTBLUE + f"<{self.annotation}>" + ENDC
            else:
                return str(self.repr_value)
        # x = repr(self.to_list())
        # if len(x) < 80:
        #     return x
        o = ""
        if self.annotation:
            o += f"/* {self.annotation} */ \n"
        if self.repr_show_gas and self.gas:
            o += OKBLUE + "{" + ENDC + str(self.gas) + OKBLUE + "} " + ENDC  # add gas for info.
        o += "[" + self._colorise_keywords(self.repr_value)
        prev_lineno = self.ast_source.lineno if self.ast_source else None
        arg_lineno = None
        annotated = False
        has_inner_newlines = False
        for arg in self.args:
            o += ",\n  "
            arg_lineno = arg.ast_source.lineno if arg.ast_source else None
            if arg_lineno is not None and arg_lineno != prev_lineno and self.value in ("seq", "if"):
                o += f"# Line {(arg_lineno)}\n  "
                prev_lineno = arg_lineno
                annotated = True
            arg_repr = arg.repr()
            if "\n" in arg_repr:
                has_inner_newlines = True
            sub = arg_repr.replace("\n", "\n  ").strip(" ")
            o += self._colorise_keywords(sub)
        output = o.rstrip(" ") + "]"
        output_on_one_line = re.sub(r",\n *", ", ", output).replace("\n", "")

        should_output_single_line = (
            (len(output_on_one_line) < 80 or len(self.args) == 1) and not annotated
        ) and not has_inner_newlines

        if should_output_single_line:
            return output_on_one_line
        else:
            return output

    def __repr__(self):
        return self.repr()

    @classmethod
    def from_list(
        cls,
        obj: Any,
        typ: VyperType = None,
        location: Optional[AddrSpace] = None,
        ast_source: Optional[vy_ast.VyperNode] = None,
        annotation: Optional[str] = None,
        error_msg: Optional[str] = None,
        mutable: bool = True,
        add_gas_estimate: int = 0,
        is_self_call: bool = False,
        passthrough_metadata: dict[str, Any] = None,
        encoding: Encoding = Encoding.VYPER,
    ) -> "IRnode":
        if isinstance(typ, str):  # pragma: nocover
            raise CompilerPanic(f"Expected type, not string: {typ}")

        if isinstance(obj, IRnode):
            # note: this modify-and-returnclause is a little weird since
            # the input gets modified. CC 20191121.
            if typ is not None:
                obj.typ = typ
            if obj.ast_source is None:
                obj.ast_source = ast_source
            if obj.location is None:
                obj.location = location
            if obj.encoding is None:
                obj.encoding = encoding
            if obj.error_msg is None:
                obj.error_msg = error_msg

            return obj
        elif not isinstance(obj, list):
            return cls(
                obj,
                [],
                typ,
                location=location,
                annotation=annotation,
                mutable=mutable,
                add_gas_estimate=add_gas_estimate,
                ast_source=ast_source,
                encoding=encoding,
                error_msg=error_msg,
                is_self_call=is_self_call,
                passthrough_metadata=passthrough_metadata,
            )
        else:
            return cls(
                obj[0],
                [cls.from_list(o, ast_source=ast_source) for o in obj[1:]],
                typ,
                location=location,
                annotation=annotation,
                mutable=mutable,
                ast_source=ast_source,
                add_gas_estimate=add_gas_estimate,
                encoding=encoding,
                error_msg=error_msg,
                is_self_call=is_self_call,
                passthrough_metadata=passthrough_metadata,
            )



import pytest

from tests.utils import decimal_to_int, parse_and_fold


@pytest.mark.parametrize("typ_name", ["decimal"])
def test_epsilon(get_contract, typ_name):
    source = f"""
@external
def foo() -> {typ_name}:
    return epsilon({typ_name})
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"epsilon({typ_name})")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo() == decimal_to_int(new_node.value)



import pytest

from vyper import compile_code
from vyper.exceptions import StructureException

fail_list = [
    (
        """
foo: DynArray[HashMap[uint8, uint8], 2]
    """,
        StructureException,
    ),
    (
        """
foo: public(DynArray[HashMap[uint8, uint8], 2])
    """,
        StructureException,
    ),
    (
        """
@external
def foo():
    a: DynArray = [1, 2, 3]
    """,
        StructureException,
    ),
    (
        """
@external
def foo():
    a: DynArray[uint256, FOO] = [1, 2, 3]
    """,
        StructureException,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_block_fail(bad_code, exc):
    with pytest.raises(exc):
        compile_code(bad_code)


valid_list = [
    """
flag Foo:
    FE
    FI

bar: DynArray[Foo, 10]
    """,  # dynamic arrays of flags are allowed, but not static arrays
    """
bar: DynArray[Bytes[30], 10]
    """,  # dynamic arrays of bytestrings are allowed, but not static arrays
]


@pytest.mark.parametrize("good_code", valid_list)
def test_dynarray_pass(good_code):
    assert compile_code(good_code) is not None



from vyper.utils import OrderedSet
from vyper.venom.analysis.analysis import IRAnalysis
from vyper.venom.basicblock import BB_TERMINATORS, CFG_ALTERING_INSTRUCTIONS


class CFGAnalysis(IRAnalysis):
    """
    Compute control flow graph information for each basic block in the function.
    """

    def analyze(self) -> None:
        fn = self.function
        for bb in fn.get_basic_blocks():
            bb.cfg_in = OrderedSet()
            bb.cfg_out = OrderedSet()
            bb.out_vars = OrderedSet()

        for bb in fn.get_basic_blocks():
            assert len(bb.instructions) > 0, "Basic block should not be empty"
            last_inst = bb.instructions[-1]
            assert (
                last_inst.opcode in BB_TERMINATORS
            ), f"Last instruction should be a terminator {bb}"

            for inst in bb.instructions:
                if inst.opcode in CFG_ALTERING_INSTRUCTIONS:
                    ops = inst.get_label_operands()
                    for op in ops:
                        fn.get_basic_block(op.value).add_cfg_in(bb)

        # Fill in the "out" set for each basic block
        for bb in fn.get_basic_blocks():
            for in_bb in bb.cfg_in:
                in_bb.add_cfg_out(bb)

    def invalidate(self):
        from vyper.venom.analysis.dominators import DominatorTreeAnalysis
        from vyper.venom.analysis.liveness import LivenessAnalysis

        self.analyses_cache.invalidate_analysis(DominatorTreeAnalysis)
        self.analyses_cache.invalidate_analysis(LivenessAnalysis)



from typing import Any, Optional

from vyper.codegen.abi_encoder import abi_encode, abi_encoding_matches_vyper
from vyper.codegen.context import Context
from vyper.codegen.core import (
    calculate_type_for_external_return,
    check_assign,
    dummy_node_for_type,
    get_type_for_exact_size,
    make_setter,
    needs_clamp,
    wrap_value_for_external_return,
)
from vyper.codegen.ir_node import IRnode
from vyper.evm.address_space import MEMORY
from vyper.exceptions import TypeCheckFailure

Stmt = Any  # mypy kludge


# Generate code for return stmt
def make_return_stmt(ir_val: IRnode, stmt: Any, context: Context) -> Optional[IRnode]:
    func_t = context.func_t

    jump_to_exit = ["exit_to", func_t._ir_info.exit_sequence_label]

    if context.return_type is None:
        if stmt.value is not None:  # pragma: nocover
            raise TypeCheckFailure("bad return")

    else:
        # sanity typecheck
        check_assign(dummy_node_for_type(context.return_type), ir_val)

    # helper function
    # do NOT bypass this. jump_to_exit may do important function cleanup.
    def finalize(fill_return_buffer):
        fill_return_buffer = IRnode.from_list(
            fill_return_buffer, annotation=f"fill return buffer {func_t._ir_info.ir_identifier}"
        )
        cleanup_loops = "cleanup_repeat" if context.forvars else "seq"
        # NOTE: because stack analysis is incomplete, cleanup_repeat must
        # come after fill_return_buffer otherwise the stack will break
        jump_to_exit_ir = IRnode.from_list(jump_to_exit)
        return IRnode.from_list(["seq", fill_return_buffer, cleanup_loops, jump_to_exit_ir])

    if context.return_type is None:
        if context.is_internal:
            jump_to_exit += ["return_pc"]
        return finalize(["seq"])

    if context.is_internal:
        dst = IRnode.from_list(["return_buffer"], typ=context.return_type, location=MEMORY)
        fill_return_buffer = make_setter(dst, ir_val)
        jump_to_exit += ["return_pc"]

        return finalize(fill_return_buffer)

    else:  # return from external function
        external_return_type = calculate_type_for_external_return(context.return_type)
        maxlen = external_return_type.abi_type.size_bound()

        # optimize: if the value already happens to be ABI encoded in
        # memory, don't bother running abi_encode, just return the
        # buffer it is in.
        can_skip_encode = (
            abi_encoding_matches_vyper(ir_val.typ)
            and ir_val.location == MEMORY
            # ensure it has already been validated - could be
            # unvalidated ABI encoded returndata for example
            and not needs_clamp(ir_val.typ, ir_val.encoding)
        )

        if can_skip_encode:
            assert ir_val.typ.memory_bytes_required == maxlen  # type: ignore
            jump_to_exit += [ir_val, maxlen]  # type: ignore
            return finalize(["pass"])

        ir_val = wrap_value_for_external_return(ir_val)

        # general case: abi_encode the data to a newly allocated buffer
        # and return the buffer
        return_buffer_ofst = context.new_internal_variable(get_type_for_exact_size(maxlen))

        # encode_out is cleverly a sequence which does the abi-encoding and
        # also returns the length of the output as a stack element
        return_len = abi_encode(return_buffer_ofst, ir_val, context, returns_len=True, bufsz=maxlen)

        # append ofst and len to exit_to the cleanup subroutine
        jump_to_exit += [return_buffer_ofst, return_len]  # type: ignore

        return finalize(["pass"])



import decimal
import math
from typing import Tuple

from vyper.codegen.core import (
    clamp,
    clamp_basetype,
    is_decimal_type,
    is_integer_type,
    is_numeric_type,
)
from vyper.codegen.ir_node import IRnode
from vyper.exceptions import CompilerPanic, TypeCheckFailure, UnimplementedException


def calculate_largest_power(a: int, num_bits: int, is_signed: bool) -> int:
    """
    For a given base `a`, compute the maximum power `b` that will not
    produce an overflow in the equation `a ** b`

    Arguments
    ---------
    a : int
        Base value for the equation `a ** b`
    num_bits : int
        The maximum number of bits that the resulting value must fit in
    is_signed : bool
        Is the operation being performed on signed integers?

    Returns
    -------
    int
        Largest possible value for `b` where the result does not overflow
        `num_bits`
    """
    if num_bits % 8:  # pragma: no cover
        raise CompilerPanic("Type is not a modulo of 8")

    if a in (-1, 0, 1):  # pragma: no cover
        raise CompilerPanic("Exponential operation is useless!")

    value_bits = num_bits - (1 if is_signed else 0)
    if a >= 2**value_bits:  # pragma: no cover
        raise TypeCheckFailure("Value is too large and will always throw")
    if a < -(2**value_bits):  # pragma: no cover
        raise TypeCheckFailure("Value is too small and will always throw")

    a_is_negative = a < 0

    a = abs(a)  # No longer need to know if it's signed or not

    # NOTE: There is an edge case if `a` were left signed where the following
    #       operation would not work (`ln(a)` is undefined if `a <= 0`)
    b = int(decimal.Decimal(value_bits) / (decimal.Decimal(a).ln() / decimal.Decimal(2).ln()))
    if b <= 1:
        return 1  # Value is assumed to be in range, therefore power of 1 is max

    # Do a bit of iteration to ensure we have the exact number

    # CMC 2022-05-06 (TODO we should be able to this with algebra
    # instead of looping):
    # a ** x == 2**value_bits
    # x ln(a) = ln(2**value_bits)
    # x = ln(2**value_bits) / ln(a)

    num_iterations = 0
    while a ** (b + 1) < 2**value_bits:
        b += 1
        num_iterations += 1
        assert num_iterations < 10000
    while a**b >= 2**value_bits:
        b -= 1
        num_iterations += 1
        assert num_iterations < 10000

    # Edge case: If a is negative and the values of a and b are such that:
    #   (-a) ** (b + 1) == -(2 ** value_bits)
    # we can squeak one more out of it because lower bound of signed ints
    # is slightly wider than upper bound
    if a_is_negative and (-a) ** (b + 1) == -(2**value_bits):  # NOTE: a = abs(a)
        return b + 1
    else:
        return b  # Exact


def calculate_largest_base(b: int, num_bits: int, is_signed: bool) -> Tuple[int, int]:
    """
    For a given power `b`, compute the maximum base `a` that will not produce an
    overflow in the equation `a ** b`

    Arguments
    ---------
    b : int
        Power value for the equation `a ** b`
    num_bits : int
        The maximum number of bits that the resulting value must fit in
    is_signed : bool
        Is the operation being performed on signed integers?

    Returns
    -------
    Tuple[int, int]
        Smallest and largest possible values for `a` where the result
        does not overflow `num_bits`.

        Note that the lower and upper bounds are not always negatives of
        each other, due to lower/upper bounds for int_<value_bits> being
        slightly asymmetric.
    """
    if num_bits % 8:  # pragma: no cover
        raise CompilerPanic("Type is not a modulo of 8")

    if b in (0, 1):  # pragma: no cover
        raise CompilerPanic("Exponential operation is useless!")

    if b < 0:  # pragma: no cover
        raise TypeCheckFailure("Cannot calculate negative exponents")

    value_bits = num_bits - (1 if is_signed else 0)
    if b > value_bits:  # pragma: no cover
        raise TypeCheckFailure("Value is too large and will always throw")

    # CMC 2022-05-06 TODO we should be able to do this with algebra
    # instead of looping):
    # x ** b == 2**value_bits
    # b ln(x) == ln(2**value_bits)
    # ln(x) == ln(2**value_bits) / b
    # x == exp( ln(2**value_bits) / b)

    # Estimate (up to ~39 digits precision required)
    a = math.ceil(2 ** (decimal.Decimal(value_bits) / decimal.Decimal(b)))
    # Do a bit of iteration to ensure we have the exact number
    num_iterations = 0
    while (a + 1) ** b < 2**value_bits:
        a += 1
        num_iterations += 1
        assert num_iterations < 10000
    while a**b >= 2**value_bits:
        a -= 1
        num_iterations += 1
        assert num_iterations < 10000

    if not is_signed:
        return 0, a

    if (a + 1) ** b == (2**value_bits):
        # edge case: lower bound is slightly wider than upper bound
        return -(a + 1), a
    else:
        return -a, a


# def safe_add(x: IRnode, y: IRnode) -> IRnode:
def safe_add(x, y):
    assert x.typ is not None and x.typ == y.typ and is_numeric_type(x.typ)
    typ = x.typ

    res = IRnode.from_list(["add", x, y], typ=typ)

    if typ.bits < 256:
        return clamp_basetype(res)

    # bits == 256
    with res.cache_when_complex("ans") as (b1, res):
        if typ.is_signed:
            # if r < 0:
            #   ans < l
            # else:
            #   ans >= l  # aka (iszero (ans < l))
            # aka: (r < 0) == (ans < l)
            ok = ["eq", ["slt", y, 0], ["slt", res, x]]
        else:
            # note this is "equivalent" to the unsigned form
            # of the above (because y < 0 == False)
            #       ["eq", ["lt", y, 0], ["lt", res, x]]
            # TODO push down into optimizer rules.
            ok = ["ge", res, x]

        check = IRnode.from_list(["assert", ok], error_msg="safeadd")
        ret = IRnode.from_list(["seq", check, res])
        return b1.resolve(ret)


# def safe_sub(x: IRnode, y: IRnode) -> IRnode:
def safe_sub(x, y):
    assert x.typ == y.typ
    typ = x.typ

    res = IRnode.from_list(["sub", x, y], typ=typ)

    if typ.bits < 256:
        return clamp_basetype(res)

    # bits == 256
    with res.cache_when_complex("ans") as (b1, res):
        if typ.is_signed:
            # if r < 0:
            #   ans > l
            # else:
            #   ans <= l  # aka (iszero (ans > l))
            # aka: (r < 0) == (ans > l)
            ok = ["eq", ["slt", y, 0], ["sgt", res, x]]
        else:
            # note this is "equivalent" to the unsigned form
            # of the above (because y < 0 == False)
            #       ["eq", ["lt", y, 0], ["gt", res, x]]
            # TODO push down into optimizer rules.
            ok = ["le", res, x]

        check = IRnode.from_list(["assert", ok], error_msg="safesub")
        ret = IRnode.from_list(["seq", check, res])
        return b1.resolve(ret)


# def safe_mul(x: IRnode, y: IRnode) -> IRnode:
def safe_mul(x, y):
    # precondition: x.typ == y.typ
    assert x.typ == y.typ
    typ = x.typ

    # optimizer rules work better for the safemul checks below
    # if second operand is literal
    if x.is_literal:
        tmp = x
        x = y
        y = tmp

    res = IRnode.from_list(["mul", x, y], typ=x.typ)

    DIV = "sdiv" if typ.is_signed else "div"

    with res.cache_when_complex("ans") as (b1, res):
        ok = [1]  # True

        if typ.bits > 128:  # check overflow mod 256
            # assert (res/y == x | y == 0)
            ok = ["or", ["eq", [DIV, res, y], x], ["iszero", y]]

        # int256
        if typ.is_signed and typ.bits == 256:
            # special case:
            # in the above sdiv check, if (r==-1 and l==-2**255),
            # -2**255<res> / -1<r> will return -2**255<l>.
            # need to check: not (r == -1 and l == -2**255)
            upper_bound = ["shl", 255, 1]

            check_x = ["ne", x, upper_bound]
            check_y = ["ne", ["not", y], 0]

            if not x.is_literal and not y.is_literal:
                # TODO can simplify this condition?
                ok = ["and", ok, ["or", check_x, check_y]]

            # TODO push some of this constant folding into optimizer
            elif x.is_literal and x.value == -(2**255):
                ok = ["and", ok, check_y]
            elif y.is_literal and y.value == -1:
                ok = ["and", ok, check_x]
            else:
                # x or y is a literal, and we have determined it is
                # not an evil value
                pass

        if is_decimal_type(res.typ):
            res = IRnode.from_list([DIV, res, typ.divisor], typ=res.typ)

        # check overflow mod <bits>
        # NOTE: if 128 < bits < 256, `x * y` could be between
        # MAX_<bits> and 2**256 OR it could overflow past 2**256.
        # so, we check for overflow in mod 256 AS WELL AS mod <bits>
        # (if bits == 256, clamp_basetype is a no-op)
        res = clamp_basetype(res)

        check = IRnode.from_list(["assert", ok], error_msg="safemul")
        res = IRnode.from_list(["seq", check, res], typ=res.typ)

        return b1.resolve(res)


# def safe_div(x: IRnode, y: IRnode) -> IRnode:
def safe_div(x, y):
    assert x.typ == y.typ
    typ = x.typ

    ok = [1]  # true

    if is_decimal_type(x.typ):
        lo, hi = typ.int_bounds
        if max(abs(lo), abs(hi)) * typ.divisor > 2**256 - 1:  # pragma: nocover
            # stub to prevent us from adding fixed point numbers we don't know
            # how to deal with
            raise UnimplementedException(f"safe_mul for decimal{typ.bits}x{typ.decimals}")
        x = ["mul", x, typ.divisor]

    DIV = "sdiv" if typ.is_signed else "div"
    res = IRnode.from_list([DIV, x, clamp("gt", y, 0)], typ=typ)
    with res.cache_when_complex("res") as (b1, res):
        # TODO: refactor this condition / push some things into the optimizer
        if typ.is_signed and typ.bits == 256:
            upper_bound = ["shl", 255, 1]

            if not x.is_literal and not y.is_literal:
                ok = ["or", ["ne", y, ["not", 0]], ["ne", x, upper_bound]]
            # TODO push these rules into the optimizer
            elif x.is_literal and x.value == -(2**255):
                ok = ["ne", y, ["not", 0]]
            elif y.is_literal and y.value == -1:
                ok = ["ne", x, upper_bound]
            else:
                # x or y is a literal, and not an evil value.
                pass

        elif typ.is_signed and is_integer_type(typ):
            lo, hi = typ.int_bounds
            # we need to throw on min_value(typ) / -1,
            # but we can skip if one of the operands is a literal and not
            # the evil value
            can_skip_clamp = (x.is_literal and x.value != lo) or (y.is_literal and y.value != -1)
            if not can_skip_clamp:
                # clamp_basetype has fewer ops than the int256 rule.
                res = clamp_basetype(res)

        elif is_decimal_type(typ):
            # always clamp decimals, since decimal division can actually
            # result in something larger than either operand (e.g. 1.0 / 0.1)
            # TODO maybe use safe_mul
            res = clamp_basetype(res)

        check = IRnode.from_list(["assert", ok], error_msg="safediv")
        return IRnode.from_list(b1.resolve(["seq", check, res]))


# def safe_mod(x: IRnode, y: IRnode) -> IRnode:
def safe_mod(x, y):
    typ = x.typ
    MOD = "smod" if typ.is_signed else "mod"
    return IRnode.from_list([MOD, x, clamp("gt", y, 0)], error_msg="safemod")


# def safe_pow(x: IRnode, y: IRnode) -> IRnode:
def safe_pow(x, y):
    typ = x.typ
    if not is_integer_type(x.typ):  # pragma: nocover
        raise TypeCheckFailure("non-integer pow")

    GE = "sge" if typ.is_signed else "ge"

    if x.is_literal:
        # cannot pass -1, 0 or 1 to `calculate_largest_power`
        if x.value in (-1, 0, 1):
            # not strictly needed, but consistent with other bases
            # (note: unsigned (ge y 0) will get optimized out)
            ok = [GE, y, 0]
        else:
            upper_bound = calculate_largest_power(x.value, typ.bits, typ.is_signed)
            # for signed integers, this also prevents negative values
            ok = ["le", y, upper_bound]

    elif y.is_literal:
        # cannot pass 0 or 1 to `calculate_largest_base`
        if y.value in (0, 1):
            ok = [1]
        else:
            lower_bound, upper_bound = calculate_largest_base(y.value, typ.bits, typ.is_signed)
            if typ.is_signed:
                ok = ["and", ["sge", x, lower_bound], ["sle", x, upper_bound]]
            else:
                ok = ["le", x, upper_bound]
    else:
        # `a ** b` where neither `a` or `b` are known
        # TODO this is currently unreachable, once we implement a way to do it safely
        # remove the check in `vyper/context/types/value/numeric.py`
        return

    assertion = IRnode.from_list(["assert", ok], error_msg="safepow")
    return IRnode.from_list(["seq", assertion, ["exp", x, y]])



import pytest

from vyper import compiler
from vyper.exceptions import TypeMismatch

fail_list = [
    (
        """
@external
def foo(x: Bytes[1]) -> Bytes[64]:
    return _abi_encode(x)
    """,
        TypeMismatch,  # output type too small
    ),
    (
        """
@external
def foo(x: Bytes[1]) -> Bytes[32]:
    return _abi_encode(x, ensure_tuple=False)
    """,
        TypeMismatch,  # output type too small
    ),
    (
        """
@external
def foo(x: uint256) -> Bytes[36]:
    _ensure_tuple: bool = False
    _method_id: Bytes[4] = method_id("foo()", output_type=Bytes[4])
    return _abi_encode(x, ensure_tuple=_ensure_tuple, method_id=_method_id)
    """,
        TypeMismatch,  # ensure_tuple and method_id both must be literals
    ),
    (
        """
@external
def foo(x: uint256) -> Bytes[36]:
    return _abi_encode(x, method_id=b"abcde")
    """,
        TypeMismatch,  # len(method_id) must be less than 4
    ),
    (
        """
@external
def foo(x: uint256) -> Bytes[36]:
    return _abi_encode(x, method_id=0x1234567890)
    """,
        TypeMismatch,  # len(method_id) must be less than 4
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_abi_encode_fail(bad_code, exc):
    with pytest.raises(exc):
        compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo(x: uint256) -> Bytes[32]:
    return _abi_encode(x)
    """,
    """
@external
def foo(x: Bytes[1]) -> Bytes[96]:
    return _abi_encode(x)
    """,
    """
@external
def foo(x: Bytes[1]) -> Bytes[64]:
    return _abi_encode(x, ensure_tuple=False)
    """,
    """
@external
def foo(x: Bytes[1]) -> Bytes[68]:
    return _abi_encode(x, ensure_tuple=False, method_id=method_id("exec()"))
    """,
    """
@external
def foo(x: Bytes[1]) -> Bytes[68]:
    return _abi_encode(x, ensure_tuple=False, method_id=0x12345678)
    """,
    """
BAR: constant(DynArray[uint256, 5]) = [1, 2, 3, 4, 5]

@external
def foo() -> Bytes[224]:
    return _abi_encode(BAR)
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_abi_encode_success(good_code):
    assert compiler.compile_code(good_code) is not None



import enum
import io
import re
from collections import defaultdict
from tokenize import COMMENT, NAME, OP, TokenError, TokenInfo, tokenize, untokenize

from packaging.specifiers import InvalidSpecifier, SpecifierSet

from vyper.compiler.settings import OptimizationLevel, Settings

# seems a bit early to be importing this but we want it to validate the
# evm-version pragma
from vyper.evm.opcodes import EVM_VERSIONS
from vyper.exceptions import StructureException, SyntaxException, VersionException
from vyper.typing import ModificationOffsets, ParserPosition


def validate_version_pragma(version_str: str, full_source_code: str, start: ParserPosition) -> None:
    """
    Validates a version pragma directive against the current compiler version.
    """
    from vyper import __version__

    if len(version_str) == 0:
        raise VersionException("Version specification cannot be empty", full_source_code, *start)

    # X.Y.Z or vX.Y.Z => ==X.Y.Z, ==vX.Y.Z
    if re.match("[v0-9]", version_str):
        version_str = "==" + version_str
    # convert npm to pep440
    version_str = re.sub("^\\^", "~=", version_str)

    try:
        spec = SpecifierSet(version_str)
    except InvalidSpecifier:
        raise VersionException(
            f'Version specification "{version_str}" is not a valid PEP440 specifier',
            full_source_code,
            *start,
        )

    if not spec.contains(__version__, prereleases=True):
        raise VersionException(
            f'Version specification "{version_str}" is not compatible '
            f'with compiler version "{__version__}"',
            full_source_code,
            *start,
        )


class ForParserState(enum.Enum):
    NOT_RUNNING = enum.auto()
    START_SOON = enum.auto()
    RUNNING = enum.auto()


# a simple state machine which allows us to handle loop variable annotations
# (which are rejected by the python parser due to pep-526, so we scoop up the
# tokens between `:` and `in` and parse them and add them back in later).
class ForParser:
    def __init__(self, code):
        self._code = code
        self.annotations = {}
        self._current_annotation = None

        self._state = ForParserState.NOT_RUNNING
        self._current_for_loop = None

    def consume(self, token):
        # state machine: we can start slurping tokens soon
        if token.type == NAME and token.string == "for":
            # note: self._state should be NOT_RUNNING here, but we don't sanity
            # check here as that should be an error the parser will handle.
            self._state = ForParserState.START_SOON
            self._current_for_loop = token.start

        if self._state == ForParserState.NOT_RUNNING:
            return False

        # state machine: start slurping tokens
        if token.type == OP and token.string == ":":
            self._state = ForParserState.RUNNING

            # sanity check -- this should never really happen, but if it does,
            # try to raise an exception which pinpoints the source.
            if self._current_annotation is not None:
                raise SyntaxException(
                    "for loop parse error", self._code, token.start[0], token.start[1]
                )

            self._current_annotation = []
            return True  # do not add ":" to tokens.

        # state machine: end slurping tokens
        if token.type == NAME and token.string == "in":
            self._state = ForParserState.NOT_RUNNING
            self.annotations[self._current_for_loop] = self._current_annotation or []
            self._current_annotation = None
            return False

        if self._state != ForParserState.RUNNING:
            return False

        # slurp the token
        self._current_annotation.append(token)
        return True


# compound statements that are replaced with `class`
# TODO remove enum in favor of flag
VYPER_CLASS_TYPES = {
    "flag": "FlagDef",
    "enum": "EnumDef",
    "event": "EventDef",
    "interface": "InterfaceDef",
    "struct": "StructDef",
}

# simple statements that are replaced with `yield`
CUSTOM_STATEMENT_TYPES = {"log": "Log"}
# expression types that are replaced with `await`
CUSTOM_EXPRESSION_TYPES = {"extcall": "ExtCall", "staticcall": "StaticCall"}


def pre_parse(code: str) -> tuple[Settings, ModificationOffsets, dict, str]:
    """
    Re-formats a vyper source string into a python source string and performs
    some validation.  More specifically,

    * Translates "interface", "struct", "flag", and "event" keywords into python "class" keyword
    * Validates "@version" pragma against current compiler version
    * Prevents direct use of python "class" keyword
    * Prevents use of python semi-colon statement separator
    * Extracts type annotation of for loop iterators into a separate dictionary

    Also returns a mapping of detected interface and struct names to their
    respective vyper class types ("interface" or "struct"), and a mapping of line numbers
    of for loops to the type annotation of their iterators.

    Parameters
    ----------
    code : str
        The vyper source code to be re-formatted.

    Returns
    -------
    Settings
        Compilation settings based on the directives in the source code
    ModificationOffsets
        A mapping of class names to their original class types.
    dict[tuple[int, int], list[TokenInfo]]
        A mapping of line/column offsets of `For` nodes to the annotation of the for loop target
    str
        Reformatted python source string.
    """
    result = []
    modification_offsets: ModificationOffsets = {}
    settings = Settings()
    for_parser = ForParser(code)

    _col_adjustments: dict[int, int] = defaultdict(lambda: 0)

    try:
        code_bytes = code.encode("utf-8")
        token_list = list(tokenize(io.BytesIO(code_bytes).readline))

        for token in token_list:
            toks = [token]

            typ = token.type
            string = token.string
            start = token.start
            end = token.end
            line = token.line

            if typ == COMMENT:
                contents = string[1:].strip()
                if contents.startswith("@version"):
                    if settings.compiler_version is not None:
                        raise StructureException("compiler version specified twice!", start)
                    compiler_version = contents.removeprefix("@version ").strip()
                    validate_version_pragma(compiler_version, code, start)
                    settings.compiler_version = compiler_version

                if contents.startswith("pragma "):
                    pragma = contents.removeprefix("pragma ").strip()
                    if pragma.startswith("version "):
                        if settings.compiler_version is not None:
                            raise StructureException("pragma version specified twice!", start)
                        compiler_version = pragma.removeprefix("version ").strip()
                        validate_version_pragma(compiler_version, code, start)
                        settings.compiler_version = compiler_version

                    # TODO: refactor these to something like Settings.from_pragma
                    elif pragma.startswith("optimize "):
                        if settings.optimize is not None:
                            raise StructureException("pragma optimize specified twice!", start)
                        try:
                            mode = pragma.removeprefix("optimize").strip()
                            settings.optimize = OptimizationLevel.from_string(mode)
                        except ValueError:
                            raise StructureException(f"Invalid optimization mode `{mode}`", start)
                    elif pragma.startswith("evm-version "):
                        if settings.evm_version is not None:
                            raise StructureException("pragma evm-version specified twice!", start)
                        evm_version = pragma.removeprefix("evm-version").strip()
                        if evm_version not in EVM_VERSIONS:
                            raise StructureException(f"Invalid evm version: `{evm_version}`", start)
                        settings.evm_version = evm_version
                    elif pragma.startswith("experimental-codegen"):
                        if settings.experimental_codegen is not None:
                            raise StructureException(
                                "pragma experimental-codegen specified twice!", start
                            )
                        settings.experimental_codegen = True
                    elif pragma.startswith("enable-decimals"):
                        if settings.enable_decimals is not None:
                            raise StructureException(
                                "pragma enable_decimals specified twice!", start
                            )
                        settings.enable_decimals = True

                    else:
                        raise StructureException(f"Unknown pragma `{pragma.split()[0]}`")

            if typ == NAME and string in ("class", "yield"):
                raise SyntaxException(
                    f"The `{string}` keyword is not allowed. ", code, start[0], start[1]
                )

            if typ == NAME:
                if string in VYPER_CLASS_TYPES and start[1] == 0:
                    toks = [TokenInfo(NAME, "class", start, end, line)]
                    modification_offsets[start] = VYPER_CLASS_TYPES[string]
                elif string in CUSTOM_STATEMENT_TYPES:
                    new_keyword = "yield"
                    adjustment = len(new_keyword) - len(string)
                    # adjustments for following staticcall/extcall modification_offsets
                    _col_adjustments[start[0]] += adjustment
                    toks = [TokenInfo(NAME, new_keyword, start, end, line)]
                    modification_offsets[start] = CUSTOM_STATEMENT_TYPES[string]
                elif string in CUSTOM_EXPRESSION_TYPES:
                    # a bit cursed technique to get untokenize to put
                    # the new tokens in the right place so that modification_offsets
                    # will work correctly.
                    # (recommend comparing the result of pre_parse with the
                    # source code side by side to visualize the whitespace)
                    new_keyword = "await"
                    vyper_type = CUSTOM_EXPRESSION_TYPES[string]

                    lineno, col_offset = start

                    # fixup for when `extcall/staticcall` follows `log`
                    adjustment = _col_adjustments[lineno]
                    new_start = (lineno, col_offset + adjustment)
                    modification_offsets[new_start] = vyper_type

                    # tells untokenize to add whitespace, preserving locations
                    diff = len(new_keyword) - len(string)
                    new_end = end[0], end[1] + diff

                    toks = [TokenInfo(NAME, new_keyword, start, new_end, line)]

            if (typ, string) == (OP, ";"):
                raise SyntaxException("Semi-colon statements not allowed", code, start[0], start[1])

            if not for_parser.consume(token):
                result.extend(toks)

    except TokenError as e:
        raise SyntaxException(e.args[0], code, e.args[1][0], e.args[1][1]) from e

    for_loop_annotations = {}
    for k, v in for_parser.annotations.items():
        for_loop_annotations[k] = v.copy()

    return settings, modification_offsets, for_loop_annotations, untokenize(result).decode("utf-8")



import pytest
from pytest import raises

from vyper import compiler
from vyper.exceptions import InvalidOperation, SyntaxException, TypeMismatch

fail_list = [
    """
@external
def foo():
    x: bool = True
    x = 5
    """,
    (
        """
@external
def foo():
    True = 3
    """,
        SyntaxException,
    ),
    """
@external
def foo():
    x: bool = True
    x = 129
    """,
    (
        """
@external
def foo() -> bool:
    return (1 == 2) <= (1 == 1)
    """,
        InvalidOperation,
    ),
    """
@external
def foo() -> bool:
    return (1 == 2) or 3
    """,
    """
@external
def foo() -> bool:
    return 1.0 == 1
    """,
    """
@external
def foo() -> bool:
    a: address = empty(address)
    return a == 1
    """,
    (
        """
@external
def foo(a: address) -> bool:
    return not a
    """,
        InvalidOperation,
    ),
    """
@external
def test(a: address) -> bool:
    assert(a)
    return True
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_bool_fail(bad_code):
    if isinstance(bad_code, tuple):
        with raises(bad_code[1]):
            compiler.compile_code(bad_code[0])
    else:
        with raises(TypeMismatch):
            compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo():
    x: bool = True
    z: bool = x and False
    """,
    """
@external
def foo():
    x: bool = True
    z: bool = x and False
    """,
    """
@external
def foo():
    x: bool = True
    x = False
    """,
    """
@external
def foo() -> bool:
    return 1 == 1
    """,
    """
@external
def foo() -> bool:
    return 1 != 1
    """,
    """
@external
def foo() -> bool:
    return 1 > 1
    """,
    """
@external
def foo() -> bool:
    return 2 >= 1
    """,
    """
@external
def foo() -> bool:
    return 1 < 1
    """,
    """
@external
def foo() -> bool:
    return 1 <= 1
    """,
    """
@external
def foo2(a: address) -> bool:
    return a != empty(address)
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_bool_success(good_code):
    assert compiler.compile_code(good_code) is not None


@pytest.mark.parametrize(
    "length,value,result",
    [
        (1, "a", False),
        (1, "", True),
        (8, "helloooo", False),
        (8, "hello", False),
        (8, "", True),
        (40, "a", False),
        (40, "hellohellohellohellohellohellohellohello", False),
        (40, "", True),
    ],
)
@pytest.mark.parametrize("op", ["==", "!="])
def test_empty_string_comparison(get_contract, length, value, result, op):
    contract = f"""
@external
def foo(xs: String[{length}]) -> bool:
    return xs {op} ""
    """
    c = get_contract(contract)
    if op == "==":
        assert c.foo(value) == result
    elif op == "!=":
        assert c.foo(value) != result


@pytest.mark.parametrize(
    "length,value,result",
    [
        (1, b"a", False),
        (1, b"", True),
        (8, b"helloooo", False),
        (8, b"hello", False),
        (8, b"", True),
        (40, b"a", False),
        (40, b"hellohellohellohellohellohellohellohello", False),
        (40, b"", True),
    ],
)
@pytest.mark.parametrize("op", ["==", "!="])
def test_empty_bytes_comparison(get_contract, length, value, result, op):
    contract = f"""
@external
def foo(xs: Bytes[{length}]) -> bool:
    return b"" {op} xs
    """
    c = get_contract(contract)
    if op == "==":
        assert c.foo(value) == result
    elif op == "!=":
        assert c.foo(value) != result



import pytest
from eth.codecs import abi
from eth_utils import to_text

from tests.utils import decimal_to_int
from vyper import compile_code
from vyper.exceptions import (
    ArgumentException,
    EventDeclarationException,
    InvalidType,
    NamespaceCollision,
    StructureException,
    TypeMismatch,
    UndeclaredDefinition,
)
from vyper.utils import keccak256


def test_empty_event_logging(get_logs, keccak, get_contract):
    loggy_code = """
event MyLog: pass

@external
def foo():
    log MyLog()
    """

    c = get_contract(loggy_code)
    c.foo()
    (log,) = get_logs(c)

    event_id = keccak(bytes("MyLog()", "utf-8"))

    # Event id is always the first topic
    assert log.topics == [event_id]
    # Event abi is created correctly
    assert c.abi[0] == {"name": "MyLog", "inputs": [], "anonymous": False, "type": "event"}
    # Event is decoded correctly
    assert log.event == "MyLog"


def test_event_logging_with_topics(get_logs, keccak, get_contract):
    loggy_code = """

a: Bytes[3]

event MyLog:
    arg1: indexed(Bytes[3])

@external
def foo():
    self.a = b"bar"
    log MyLog(self.a)
    """

    c = get_contract(loggy_code)
    c.foo()
    event_id = keccak(bytes("MyLog(bytes)", "utf-8"))

    # Event id is always the first topic
    (log,) = get_logs(c)
    assert log.topics[0] == event_id
    # Event abi is created correctly
    assert c.abi[0] == {
        "name": "MyLog",
        "inputs": [{"type": "bytes", "name": "arg1", "indexed": True}],
        "anonymous": False,
        "type": "event",
    }


def test_event_logging_with_multiple_topics(env, keccak, get_logs, get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(int128)
    arg2: indexed(bool)
    arg3: indexed(address)

@external
def foo():
    log MyLog(-2, True, self)
    """

    c = get_contract(loggy_code)
    c.foo()

    event_id = keccak(bytes("MyLog(int128,bool,address)", "utf-8"))
    # Event id is always the first topic
    (log,) = get_logs(c)
    assert log.topics[0] == event_id
    # Event abi is created correctly
    assert c.abi[0] == {
        "name": "MyLog",
        "inputs": [
            {"type": "int128", "name": "arg1", "indexed": True},
            {"type": "bool", "name": "arg2", "indexed": True},
            {"type": "address", "name": "arg3", "indexed": True},
        ],
        "anonymous": False,
        "type": "event",
    }
    # Event is decoded correctly
    (log,) = get_logs(c, "MyLog")
    assert log.event == "MyLog"
    assert log.args.arg1 == -2
    assert log.args.arg2 is True
    assert log.args.arg3 == c.address


def test_event_logging_with_multiple_topics_var_and_store(get_contract, get_logs):
    code = """
event MyLog:
    arg1: indexed(int128)
    arg2: indexed(bool)
    arg3: indexed(address)

b: address

@external
def foo(arg1: int128):
    a: bool = True
    self.b = self
    log MyLog(arg1, a, self.b)
    """

    c = get_contract(code)
    c.foo(31337)

    # Event is decoded correctly
    (log,) = get_logs(c, "MyLog")
    assert log.args.arg1 == 31337
    assert log.args.arg2 is True
    assert log.args.arg3 == c.address


def test_logging_the_same_event_multiple_times_with_topics(env, keccak, get_logs, get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(int128)
    arg2: indexed(address)

@external
def foo():
    log MyLog(1, self)
    log MyLog(1, self)

@external
def bar():
    log MyLog(1, self)
    log MyLog(1, self)
    """

    c = get_contract(loggy_code)
    c.foo()
    log1, log2 = get_logs(c)
    c.bar()
    log3, _ = get_logs(c)

    event_id = keccak(bytes("MyLog(int128,address)", "utf-8"))
    # Event id is always the first topic
    assert log1.topics[0] == event_id
    assert log3.topics[0] == event_id
    # Event abi is created correctly
    assert c.abi[0] == {
        "name": "MyLog",
        "inputs": [
            {"type": "int128", "name": "arg1", "indexed": True},
            {"type": "address", "name": "arg2", "indexed": True},
        ],
        "anonymous": False,
        "type": "event",
    }

    # Event is decoded correctly
    assert log1.args.arg1 == 1
    assert log1.args.arg2 == c.address
    assert log2.args.arg1 == 1
    assert log2.args.arg2 == c.address


def test_event_logging_cannot_have_more_than_three_topics(tx_failed, get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(Bytes[3])
    arg2: indexed(Bytes[4])
    arg3: indexed(address)
    arg4: indexed(int128)
    """

    with pytest.raises(EventDeclarationException):
        compile_code(loggy_code)


def test_event_logging_with_data(get_logs, keccak, get_contract):
    loggy_code = """
event MyLog:
    arg1: int128

@external
def foo():
    log MyLog(123)
    """

    c = get_contract(loggy_code)
    c.foo()

    event_id = keccak(bytes("MyLog(int128)", "utf-8"))
    # Event id is always the first topic
    (log,) = get_logs(c)
    assert log.topics == [event_id]
    # Event abi is created correctly
    assert c.abi[0] == {
        "name": "MyLog",
        "inputs": [{"type": "int128", "name": "arg1", "indexed": False}],
        "anonymous": False,
        "type": "event",
    }

    # Event is decoded correctly
    (log,) = get_logs(c, "MyLog")
    assert log.args.arg1 == 123


def test_event_logging_with_fixed_array_data(env, keccak, get_logs, get_contract):
    loggy_code = """
event MyLog:
    arg1: int128[2]
    arg2: uint256[3]
    arg3: int128[2][2]


@external
def foo():
    log MyLog([1,2], [block.timestamp, block.timestamp+1, block.timestamp+2], [[1,2],[1,2]])
    log MyLog([1,2], [block.timestamp, block.timestamp+1, block.timestamp+2], [[1,2],[1,2]])
    """

    c = get_contract(loggy_code)
    c.foo()

    event_id = keccak(bytes("MyLog(int128[2],uint256[3],int128[2][2])", "utf-8"))
    # Event id is always the first topic
    log, _ = get_logs(c)
    assert log.topics == [event_id]
    # Event abi is created correctly

    assert c.abi[0] == {
        "name": "MyLog",
        "inputs": [
            {"type": "int128[2]", "name": "arg1", "indexed": False},
            {"type": "uint256[3]", "name": "arg2", "indexed": False},
            {"type": "int128[2][2]", "name": "arg3", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    }

    # Event is decoded correctly
    timestamp = env.timestamp
    assert log.args.arg1 == [1, 2]
    assert log.args.arg2 == [timestamp, timestamp + 1, timestamp + 2]
    assert log.args.arg3 == [[1, 2], [1, 2]]


def test_logging_with_input_bytes_1(env, keccak, get_logs, get_contract):
    loggy_code = """
event MyLog:
    arg1: Bytes[4]
    arg2: indexed(Bytes[29])
    arg3: Bytes[31]

@external
def foo(arg1: Bytes[29], arg2: Bytes[31]):
    log MyLog(b'bar', arg1, arg2)
"""

    c = get_contract(loggy_code)
    c.foo(b"bar", b"foo")

    event_id = keccak(bytes("MyLog(bytes,bytes,bytes)", "utf-8"))
    # Event id is always the first topic
    (log,) = get_logs(c)
    assert log.topics[0] == event_id
    # Event abi is created correctly
    assert c.abi[0] == {
        "name": "MyLog",
        "inputs": [
            {"type": "bytes", "name": "arg1", "indexed": False},
            {"type": "bytes", "name": "arg2", "indexed": True},
            {"type": "bytes", "name": "arg3", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    }
    # Event is decoded correctly
    (log,) = get_logs(c, "MyLog")

    assert log.args.arg1 == b"bar"
    assert log.args.arg2 == keccak(b"bar")
    assert log.args.arg3 == b"foo"


def test_event_logging_with_bytes_input_2(env, keccak, get_logs, get_contract):
    loggy_code = """
event MyLog:
    arg1: Bytes[20]

@external
def foo(_arg1: Bytes[20]):
    log MyLog(_arg1)
    """

    c = get_contract(loggy_code)
    c.foo(b"hello")

    event_id = keccak(bytes("MyLog(bytes)", "utf-8"))
    # Event id is always the first topic
    (log,) = get_logs(c)
    assert log.topics[0] == event_id
    # Event abi is created correctly
    assert c.abi[0] == {
        "anonymous": False,
        "inputs": [{"indexed": False, "name": "arg1", "type": "bytes"}],
        "name": "MyLog",
        "type": "event",
    }
    # Event is decoded correctly
    assert log.args.arg1 == b"hello"


def test_event_logging_with_bytes_input_3(get_logs, keccak, get_contract):
    loggy_code = """
event MyLog:
    arg1: Bytes[5]

@external
def foo(_arg1: Bytes[5]):
    log MyLog(_arg1)
    """

    c = get_contract(loggy_code)
    c.foo(b"hello")

    event_id = keccak(bytes("MyLog(bytes)", "utf-8"))
    # Event id is always the first topic
    (log,) = get_logs(c)
    assert log.topics == [event_id]
    # Event abi is created correctly
    assert c.abi[0] == {
        "anonymous": False,
        "inputs": [{"indexed": False, "name": "arg1", "type": "bytes"}],
        "name": "MyLog",
        "type": "event",
    }
    # Event is decoded correctly
    (log,) = get_logs(c, "MyLog")
    assert log.args.arg1 == b"hello"


def test_event_logging_with_data_with_different_types(env, keccak, get_logs, get_contract):
    loggy_code = """
event MyLog:
    arg1: int128
    arg2: Bytes[4]
    arg3: Bytes[3]
    arg4: address
    arg5: address
    arg6: uint256

@external
def foo():
    log MyLog(123, b'home', b'bar', 0xc305c901078781C232A2a521C2aF7980f8385ee9, self, block.timestamp)  # noqa: E501
    """

    c = get_contract(loggy_code)
    c.foo()

    event_id = keccak(bytes("MyLog(int128,bytes,bytes,address,address,uint256)", "utf-8"))
    # Event id is always the first topic
    (log,) = get_logs(c)
    assert log.topics == [event_id]
    # Event abi is created correctly
    assert c.abi[0] == {
        "name": "MyLog",
        "inputs": [
            {"type": "int128", "name": "arg1", "indexed": False},
            {"type": "bytes", "name": "arg2", "indexed": False},
            {"type": "bytes", "name": "arg3", "indexed": False},
            {"type": "address", "name": "arg4", "indexed": False},
            {"type": "address", "name": "arg5", "indexed": False},
            {"type": "uint256", "name": "arg6", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    }

    # Event is decoded correctly
    args = log.args
    assert args.arg1 == 123
    assert args.arg2 == b"home"
    assert args.arg3 == b"bar"
    assert args.arg4 == "0xc305c901078781C232A2a521C2aF7980f8385ee9"
    assert args.arg5 == c.address
    assert args.arg6 == env.timestamp


def test_event_logging_with_topics_and_data_1(env, keccak, get_logs, get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(int128)
    arg2: Bytes[3]

@external
def foo():
    log MyLog(1, b'bar')
    """

    c = get_contract(loggy_code)
    c.foo()

    event_id = keccak(bytes("MyLog(int128,bytes)", "utf-8"))
    # Event id is always the first topic
    (log,) = get_logs(c)
    assert log.topics[0] == event_id
    # Event abi is created correctly
    assert c.abi[0] == {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "name": "arg1", "type": "int128"},
            {"indexed": False, "name": "arg2", "type": "bytes"},
        ],
        "name": "MyLog",
        "type": "event",
    }
    # Event is decoded correctly
    (log,) = get_logs(c, "MyLog")
    assert log.args.arg1 == 1
    assert log.args.arg2 == b"bar"


def test_event_logging_with_multiple_logs_topics_and_data(env, keccak, get_logs, get_contract):
    loggy_code = """
struct SmallStruct:
    t: String[5]
    w: decimal
struct MyStruct:
    x: uint256
    y: Bytes[3]
    z: SmallStruct

event MyLog:
    arg1: indexed(int128)
    arg2: Bytes[3]
event YourLog:
    arg1: indexed(address)
    arg2: MyStruct

@external
def foo():
    log MyLog(1, b'bar')
    log YourLog(self, MyStruct(x=1, y=b'abc', z=SmallStruct(t='house', w=13.5)))
    """

    c = get_contract(loggy_code)
    c.foo()

    log1, log2 = get_logs(c)
    event_id1 = keccak(bytes("MyLog(int128,bytes)", "utf-8"))
    event_id2 = keccak(bytes("YourLog(address,(uint256,bytes,(string,int168)))", "utf-8"))

    # Event id is always the first topic
    assert log1.topics[0] == event_id1
    assert log2.topics[0] == event_id2
    # Event abi is created correctly
    assert c.abi[0] == {
        "name": "MyLog",
        "inputs": [
            {"type": "int128", "name": "arg1", "indexed": True},
            {"type": "bytes", "name": "arg2", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    }
    assert c.abi[1] == {
        "name": "YourLog",
        "inputs": [
            {"name": "arg1", "type": "address", "indexed": True},
            {
                "name": "arg2",
                "type": "tuple",
                "components": [
                    {"name": "x", "type": "uint256"},
                    {"name": "y", "type": "bytes"},
                    {
                        "name": "z",
                        "type": "tuple",
                        "components": [
                            {"name": "t", "type": "string"},
                            {"name": "w", "type": "int168", "internalType": "decimal"},
                        ],
                    },
                ],
                "indexed": False,
            },
        ],
        "anonymous": False,
        "type": "event",
    }

    # Event is decoded correctly
    (my_log,) = get_logs(c, "MyLog")
    assert my_log.args.arg1 == 1
    assert my_log.args.arg2 == b"bar"

    (your_log,) = get_logs(c, "YourLog")
    assert your_log.args.arg1 == c.address
    assert your_log.args.arg2 == (1, b"abc", ("house", decimal_to_int("13.5")))


def test_fails_when_input_is_the_wrong_type(tx_failed, get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(int128)

@external
def foo_():
    log MyLog(b'yo')
"""

    with tx_failed(TypeMismatch):
        get_contract(loggy_code)


def test_fails_when_topic_is_the_wrong_size(tx_failed, get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(Bytes[3])


@external
def foo():
    log MyLog(b'bars')
"""

    with tx_failed(TypeMismatch):
        get_contract(loggy_code)


def test_fails_when_input_topic_is_the_wrong_size(tx_failed, get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(Bytes[3])

@external
def foo(arg1: Bytes[4]):
    log MyLog(arg1)
"""

    with tx_failed(TypeMismatch):
        get_contract(loggy_code)


def test_fails_when_data_is_the_wrong_size(tx_failed, get_contract):
    loggy_code = """
event MyLog:
    arg1: Bytes[3]

@external
def foo():
    log MyLog(b'bars')
"""

    with tx_failed(TypeMismatch):
        get_contract(loggy_code)


def test_fails_when_input_data_is_the_wrong_size(tx_failed, get_contract):
    loggy_code = """
event MyLog:
    arg1: Bytes[3]

@external
def foo(arg1: Bytes[4]):
    log MyLog(arg1)
"""

    with tx_failed(TypeMismatch):
        get_contract(loggy_code)


def test_topic_over_32_bytes(get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(Bytes[100])

@external
def foo():
    pass
    """
    get_contract(loggy_code)


def test_logging_fails_with_over_three_topics(tx_failed, get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(int128)
    arg2: indexed(int128)
    arg3: indexed(int128)
    arg4: indexed(int128)

@deploy
def __init__():
    log MyLog(1, 2, 3, 4)
    """

    with tx_failed(EventDeclarationException):
        get_contract(loggy_code)


def test_logging_fails_with_duplicate_log_names(tx_failed, get_contract):
    loggy_code = """
event MyLog: pass
event MyLog: pass

@external
def foo():
    log MyLog()
    """

    with tx_failed(NamespaceCollision):
        get_contract(loggy_code)


def test_logging_fails_with_when_log_is_undeclared(tx_failed, get_contract):
    loggy_code = """

@external
def foo():
    log MyLog()
    """

    with tx_failed(UndeclaredDefinition):
        get_contract(loggy_code)


def test_logging_fails_with_topic_type_mismatch(tx_failed, get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(int128)

@external
def foo():
    log MyLog(self)
    """

    with tx_failed(TypeMismatch):
        get_contract(loggy_code)


def test_logging_fails_with_data_type_mismatch(tx_failed, get_contract):
    loggy_code = """
event MyLog:
    arg1: Bytes[3]

@external
def foo():
    log MyLog(self)
    """

    with tx_failed(TypeMismatch):
        get_contract(loggy_code)


def test_logging_fails_when_number_of_arguments_is_greater_than_declaration(
    tx_failed, get_contract
):
    loggy_code = """
event MyLog:
    arg1: int128

@external
def foo():
    log MyLog(1, 2)
"""
    with tx_failed(ArgumentException):
        get_contract(loggy_code)


def test_logging_fails_when_number_of_arguments_is_less_than_declaration(tx_failed, get_contract):
    loggy_code = """
event MyLog:
    arg1: int128
    arg2: int128

@external
def foo():
    log MyLog(1)
"""
    with tx_failed(ArgumentException):
        get_contract(loggy_code)


def test_loggy_code(get_logs, get_contract):
    loggy_code = """
s: Bytes[100]

@external
def foo():
    raw_log([], b"moo")

@external
def goo():
    raw_log([0x1234567812345678123456781234567812345678123456781234567812345678], b"moo2")

@external
def hoo():
    self.s = b"moo3"
    raw_log([], self.s)

@external
def ioo(inp: Bytes[100]):
    raw_log([], inp)
    """

    c = get_contract(loggy_code)

    c.foo()
    ((topics, data),) = get_logs(c, raw=True)

    assert to_text(data) == "moo"
    c.goo()
    ((topics, data),) = get_logs(c, raw=True)

    assert to_text(data) == "moo2"
    assert topics[0] == bytes.fromhex(
        "1234567812345678123456781234567812345678123456781234567812345678"
    )

    c.hoo()
    ((topics, data),) = get_logs(c, raw=True)
    assert to_text(data) == "moo3"

    c.ioo(b"moo4")
    ((topics, data),) = get_logs(c, raw=True)
    assert to_text(data) == "moo4"

    print("Passed raw log tests")


def test_raw_log_topic_double_eval(get_contract, get_logs):
    t1 = "0x1111111111111111111111111111111111111111111111111111111111111111"
    code = f"""
x: bytes32
c: public(uint256)

@internal
def bar() -> bytes32:
    self.c += 1
    return {t1}

@external
def foo():
    self.x = {t1}
    raw_log([self.bar(), self.bar()], b"")

    """

    c = get_contract(code)
    c.foo()

    assert c.c() == 2


def test_raw_log_with_topics_in_storage_locs(get_contract, get_logs):
    t1 = "0x1111111111111111111111111111111111111111111111111111111111111111"
    t2 = "0x2222222222222222222222222222222222222222222222222222222222222222"
    code = f"""
x: bytes32
@external
def foo():
    self.x = {t1}
    raw_log([self.x], b"")
    y: bytes32 = {t2}
    raw_log([y], b"")
    """

    c = get_contract(code)
    c.foo()
    logs = get_logs(c, raw=True)

    assert len(logs) == 2
    assert logs[0][0][0] == int(t1, 16).to_bytes(32, "big")
    assert logs[1][0][0] == int(t2, 16).to_bytes(32, "big")


def test_raw_log_with_topics_in_storage_locs2(get_contract, get_logs):
    t1 = "0x1111111111111111111111111111111111111111111111111111111111111111"
    t2 = "0x2222222222222222222222222222222222222222222222222222222222222222"
    t3 = "0x3333333333333333333333333333333333333333333333333333333333333333"
    t4 = "0x4444444444444444444444444444444444444444444444444444444444444444"
    code = f"""
x: bytes32
x2: bytes32

@external
def foo():
    self.x = {t1}
    self.x2 = {t2}
    y: bytes32 = {t3}
    y2: bytes32 = {t4}
    raw_log([self.x, y, self.x2, y2], b"")

    raw_log([y, self.x], b"")
    """

    c = get_contract(code)
    c.foo()
    logs = get_logs(c, raw=True)

    assert len(logs) == 2
    assert logs[0][0] == [int(t, 16).to_bytes(32, "big") for t in [t1, t3, t2, t4]]
    assert logs[1][0] == [int(t, 16).to_bytes(32, "big") for t in [t3, t1]]


def test_raw_call_bytes32_data(get_logs, get_contract):
    code = """
b: uint256

@external
def foo():
    a: uint256 = 1234
    self.b = 4321
    raw_log([], convert(a, bytes32))
    raw_log([], convert(self.b, bytes32))
    raw_log([], convert(b"testmessage", bytes32))
    raw_log([], keccak256(b""))
    """
    c = get_contract(code)
    c.foo()
    logs = get_logs(c, raw=True)
    assert logs[0][1] == (1234).to_bytes(32, "big")
    assert logs[1][1] == (4321).to_bytes(32, "big")
    assert logs[2][1] == b"testmessage".ljust(32, b"\0")
    assert logs[3][1] == keccak256(b"")


def test_variable_list_packing(get_logs, get_contract):
    code = """
event Bar:
    _value: int128[4]

@external
def foo():
    a: int128[4] = [1, 2, 3, 4]
    log Bar(a)
    """
    c = get_contract(code)

    c.foo()
    (log,) = get_logs(c, "Bar")
    assert log.args._value == [1, 2, 3, 4]


def test_literal_list_packing(get_logs, get_contract):
    code = """
event Bar:
    _value: int128[4]

@external
def foo():
    log Bar([1, 2, 3, 4])
    """
    c = get_contract(code)

    c.foo()
    (log,) = get_logs(c, "Bar")
    assert log.args._value == [1, 2, 3, 4]


def test_storage_list_packing(get_logs, get_contract):
    code = """
event Bar:
    _value: int128[4]

x: int128[4]

@external
def foo():
    log Bar(self.x)

@external
def set_list():
    self.x = [1, 2, 3, 4]
    """
    c = get_contract(code)

    c.foo()
    (log,) = get_logs(c, "Bar")
    assert log.args._value == [0, 0, 0, 0]
    c.set_list()
    c.foo()
    (log,) = get_logs(c, "Bar")
    assert log.args._value == [1, 2, 3, 4]


def test_passed_list_packing(get_logs, get_contract):
    code = """
event Bar:
    _value: int128[4]

@external
def foo(barbaric: int128[4]):
    log Bar(barbaric)
    """
    c = get_contract(code)

    c.foo([4, 5, 6, 7])
    (log,) = get_logs(c, "Bar")
    assert log.args._value == [4, 5, 6, 7]


def test_variable_decimal_list_packing(get_logs, get_contract):
    code = """
event Bar:
    _value: decimal[4]

@external
def foo():
    log Bar([1.11, 2.22, 3.33, 4.44])
    """
    c = get_contract(code)

    c.foo()
    (log,) = get_logs(c, "Bar")
    assert log.args._value == [
        decimal_to_int("1.11"),
        decimal_to_int("2.22"),
        decimal_to_int("3.33"),
        decimal_to_int("4.44"),
    ]


def test_storage_byte_packing(get_logs, get_contract):
    code = """
event MyLog:
    arg1: Bytes[29]

x:Bytes[5]

@external
def foo(a: int128):
    log MyLog(self.x)

@external
def setbytez():
    self.x = b'hello'
    """

    c = get_contract(code)
    c.foo(0)
    (log,) = get_logs(c, "MyLog")
    assert log.args.arg1 == b""
    c.setbytez()
    c.foo(0)
    (log,) = get_logs(c, "MyLog")
    assert log.args.arg1 == b"hello"


def test_storage_decimal_list_packing(get_logs, get_contract):
    code = """
event Bar:
    _value: decimal[4]

x: decimal[4]

@external
def foo():
    log Bar(self.x)

@external
def set_list():
    self.x = [1.33, 2.33, 3.33, 4.33]
    """
    c = get_contract(code)

    c.foo()
    (log,) = get_logs(c, "Bar")
    assert log.args._value == [decimal_to_int("0")] * 4
    c.set_list()
    c.foo()
    (log,) = get_logs(c, "Bar")
    assert log.args._value == [
        decimal_to_int("1.33"),
        decimal_to_int("2.33"),
        decimal_to_int("3.33"),
        decimal_to_int("4.33"),
    ]


def test_logging_fails_when_input_is_too_big(tx_failed, get_contract):
    code = """
event Bar:
    _value: indexed(Bytes[32])

@external
def foo(inp: Bytes[33]):
    log Bar(inp)
"""
    with tx_failed(TypeMismatch):
        get_contract(code)


def test_2nd_var_list_packing(get_logs, get_contract):
    code = """
event Bar:
    arg1: int128
    arg2: int128[4]

@external
def foo():
    a: int128[4] = [1, 2, 3, 4]
    log Bar(10, a)
    """
    c = get_contract(code)

    c.foo()
    assert get_logs(c, "Bar")[0].args.arg2 == [1, 2, 3, 4]


def test_2nd_var_storage_list_packing(get_logs, get_contract):
    code = """
event Bar:
    arg1: int128
    arg2: int128[4]

x: int128[4]

@external
def foo():
    log Bar(10, self.x)

@external
def set_list():
    self.x = [1, 2, 3, 4]
    """
    c = get_contract(code)

    c.foo()
    assert get_logs(c, "Bar")[0].args.arg2 == [0, 0, 0, 0]
    c.set_list()
    c.foo()
    assert get_logs(c, "Bar")[0].args.arg2 == [1, 2, 3, 4]


def test_mixed_var_list_packing(get_logs, get_contract):
    code = """
event Bar:
    arg1: int128
    arg2: int128[4]
    arg3 :Bytes[4]
    arg4: int128[3]
    arg5: int128[2]

x: int128[4]
y: int128[2]

@deploy
def __init__():
    self.y = [1024, 2048]

@external
def foo():
    v: int128[3] = [7, 8, 9]
    log Bar(10, self.x, b"test", v, self.y)

@external
def set_list():
    self.x = [1, 2, 3, 4]
    """
    c = get_contract(code)

    c.foo()
    (log,) = get_logs(c, "Bar")
    assert log.args.arg2 == [0, 0, 0, 0]
    assert log.args.arg3 == b"test"
    assert log.args.arg4 == [7, 8, 9]
    assert log.args.arg5 == [1024, 2048]

    c.set_list()
    c.foo()
    (log,) = get_logs(c, "Bar")
    assert log.args.arg2 == [1, 2, 3, 4]
    assert log.args.arg3 == b"test"
    assert log.args.arg4 == [7, 8, 9]
    assert log.args.arg5 == [1024, 2048]


def test_hashed_indexed_topics_calldata(get_logs, keccak, get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(Bytes[36])
    arg2: indexed(int128)
    arg3: indexed(String[7])

@external
def foo(a: Bytes[36], b: int128, c: String[7]):
    log MyLog(a, b, c)
    """

    c = get_contract(loggy_code)
    c.foo(b"bar", 1, "weird")

    # Event id is always the first topic
    event_id = keccak(b"MyLog(bytes,int128,string)")
    topic1 = keccak256(b"bar")
    topic2 = abi.encode("int128", 1)
    topic3 = keccak256(b"weird")

    (log,) = get_logs(c)
    assert log.topics == [event_id, topic1, topic2, topic3]

    # Event abi is created correctly
    assert c.abi[0] == {
        "name": "MyLog",
        "inputs": [
            {"type": "bytes", "name": "arg1", "indexed": True},
            {"type": "int128", "name": "arg2", "indexed": True},
            {"type": "string", "name": "arg3", "indexed": True},
        ],
        "anonymous": False,
        "type": "event",
    }


def test_hashed_indexed_topics_memory(get_logs, keccak, get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(Bytes[10])
    arg2: indexed(int128)
    arg3: indexed(String[44])

@external
def foo():
    a: Bytes[10] = b"potato"
    b: int128 = -777
    c: String[44] = "why hello, neighbor! how are you today?"
    log MyLog(a, b, c)
    """

    c = get_contract(loggy_code)
    c.foo()

    # Event id is always the first topic
    event_id = keccak(b"MyLog(bytes,int128,string)")
    topic1 = keccak256(b"potato")
    topic2 = abi.encode("int128", -777)
    topic3 = keccak256(b"why hello, neighbor! how are you today?")
    (log,) = get_logs(c)
    assert log.topics == [event_id, topic1, topic2, topic3]

    # Event abi is created correctly
    assert c.abi[0] == {
        "name": "MyLog",
        "inputs": [
            {"type": "bytes", "name": "arg1", "indexed": True},
            {"type": "int128", "name": "arg2", "indexed": True},
            {"type": "string", "name": "arg3", "indexed": True},
        ],
        "anonymous": False,
        "type": "event",
    }


def test_hashed_indexed_topics_storage(get_logs, keccak, get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(Bytes[32])
    arg2: indexed(int128)
    arg3: indexed(String[6])

a: Bytes[32]
b: int128
c: String[6]


@external
def setter(_a: Bytes[32], _b: int128, _c: String[6]):
    self.a = _a
    self.b = _b
    self.c = _c

@external
def foo():
    log MyLog(self.a, self.b, self.c)
    """

    c = get_contract(loggy_code)
    c.setter(b"zonk", -2109, "yessir")
    c.foo()

    # Event id is always the first topic
    event_id = keccak(b"MyLog(bytes,int128,string)")
    topic1 = keccak256(b"zonk")
    topic2 = abi.encode("int128", -2109)
    topic3 = keccak256(b"yessir")

    (log,) = get_logs(c)
    assert log.topics == [event_id, topic1, topic2, topic3]

    # Event abi is created correctly
    assert c.abi[0] == {
        "name": "MyLog",
        "inputs": [
            {"type": "bytes", "name": "arg1", "indexed": True},
            {"type": "int128", "name": "arg2", "indexed": True},
            {"type": "string", "name": "arg3", "indexed": True},
        ],
        "anonymous": False,
        "type": "event",
    }


def test_hashed_indexed_topics_storxxage(get_logs, keccak, get_contract):
    loggy_code = """
event MyLog:
    arg1: indexed(Bytes[64])
    arg2: indexed(int128)
    arg3: indexed(String[21])

@external
def foo():
    log MyLog(b"wow", 666, "madness!")
    """

    c = get_contract(loggy_code)
    c.foo()

    # Event id is always the first topic
    event_id = keccak(b"MyLog(bytes,int128,string)")
    topic1 = keccak256(b"wow")
    topic2 = abi.encode("int128", 666)
    topic3 = keccak256(b"madness!")

    (log,) = get_logs(c)
    assert log.topics == [event_id, topic1, topic2, topic3]


fail_list = [
    (
        """
@external
def foo():
    raw_log([1, 2], b"moo")
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo():
    raw_log([1, 2], b"moo")
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo():
    a: DynArray[bytes32, 1] = [0x1234567812345678123456781234567812345678123456781234567812345678]
    raw_log(a, b"moo2")
    """,
        InvalidType,
    ),
    (
        """
@external
def foo():
    raw_log([b"cow"], b"dog")
    """,
        (StructureException, TypeMismatch),
    ),
    (
        """
@external
def foo():
    # bytes20 instead of bytes32
    raw_log([], 0x1234567890123456789012345678901234567890)
    """,
        TypeMismatch,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_raw_log_fail(get_contract, bad_code, exc, assert_compile_failed):
    assert_compile_failed(lambda: get_contract(bad_code), exc)



import math
from decimal import Decimal

from tests.utils import decimal_to_int


def test_ceil(get_contract):
    code = """
x: decimal

@deploy
def __init__():
    self.x = 504.0000000001

@external
def x_ceil() -> int256:
    return ceil(self.x)

@external
def foo() -> int256:
    return ceil(.9999999999)

@external
def fop() -> int256:
    return ceil(.0000000001)

@external
def foq() -> int256:
    return ceil(18707220957835557353007165858768422651595.9365500927)

@external
def fos() -> int256:
    return ceil(0.0)

@external
def fou() -> int256:
    a: int128 = 305
    b: int128 = 100
    c: decimal = convert(a, decimal) / convert(b, decimal)
    return ceil(c)
    """

    c = get_contract(code)

    assert c.x_ceil() == 505
    assert c.foo() == 1
    assert c.fop() == 1
    assert c.foq() == math.ceil(Decimal(2**167) / 10**10)
    assert c.fos() == 0
    assert c.fou() == 4


# ceil(x) should yield the smallest integer greater than or equal to x
def test_ceil_negative(get_contract):
    code = """
x: decimal

@deploy
def __init__():
    self.x = -504.0000000001

@external
def x_ceil() -> int256:
    return ceil(self.x)

@external
def foo() -> int256:
    return ceil(-11.01)

@external
def fop() -> int256:
    return ceil(-5.0)

@external
def foq() -> int256:
    return ceil(-.0000000001)

@external
def fos() -> int256:
    return ceil(-5472.9999999999)

@external
def fot() -> int256:
    return ceil(-18707220957835557353007165858768422651595.9365500928)

@external
def fou() -> int256:
    a: decimal = -305.0
    b: decimal = 100.0
    c: decimal = a / b
    return ceil(c)

@external
def ceil_param(p: decimal) -> int256:
    return ceil(p)
    """

    c = get_contract(code)

    assert c.x_ceil() == -504
    assert c.foo() == -11
    assert c.fop() == -5
    assert c.foq() == 0
    assert c.fos() == -5472
    assert c.fot() == math.ceil(-(Decimal(2**167 - 1)) / 10**10)
    assert c.fou() == -3
    assert c.ceil_param(decimal_to_int("-0.5")) == 0
    assert c.ceil_param(decimal_to_int("-7777777.7777777")) == -7777777


def test_ceil_ext_call(side_effects_contract, assert_side_effects_invoked, get_contract):
    code = """
interface Foo:
    def foo(x: decimal) -> decimal: payable

@external
def foo(a: Foo) -> int256:
    return ceil(extcall a.foo(2.5))
    """

    c1 = side_effects_contract("decimal")
    c2 = get_contract(code)

    assert c2.foo(c1.address) == 3

    assert_side_effects_invoked(c1, lambda: c2.foo(c1.address))


def test_ceil_internal_call(get_contract):
    code = """
@external
def foo() -> int256:
    return ceil(self.bar())

@internal
def bar() -> decimal:
    return 2.5
    """

    c = get_contract(code)

    assert c.foo() == 3



from vyper.compiler.settings import OptimizationLevel
from vyper.venom import generate_assembly_experimental
from vyper.venom.context import IRContext


def test_duplicate_operands():
    """
    Test the duplicate operands code generation.
    The venom code:

    %1 = 10
    %2 = add %1, %1
    %3 = mul %1, %2
    stop

    Should compile to: [PUSH1, 10, DUP1, DUP1, DUP1, ADD, MUL, STOP]
    """
    ctx = IRContext()
    fn = ctx.create_function("test")
    bb = fn.get_basic_block()
    op = bb.append_instruction("store", 10)
    sum_ = bb.append_instruction("add", op, op)
    bb.append_instruction("mul", sum_, op)
    bb.append_instruction("stop")

    asm = generate_assembly_experimental(ctx, optimize=OptimizationLevel.GAS)
    assert asm == ["PUSH1", 10, "DUP1", "DUP1", "ADD", "MUL", "STOP"]



import pytest

from vyper.ast import parse_to_ast
from vyper.exceptions import ArgumentException, ImmutableViolation, StructureException, TypeMismatch
from vyper.semantics.analysis import analyze_module


def test_modify_iterator_function_outside_loop(dummy_input_bundle):
    code = """

a: uint256[3]

@internal
def foo():
    self.a[0] = 1

@internal
def bar():
    self.foo()
    for i: uint256 in self.a:
        pass
    """
    vyper_module = parse_to_ast(code)
    analyze_module(vyper_module, dummy_input_bundle)


def test_pass_memory_var_to_other_function(dummy_input_bundle):
    code = """

@internal
def foo(a: uint256[3]) -> uint256[3]:
    b: uint256[3] = a
    b[1] = 42
    return b


@internal
def bar():
    a: uint256[3] = [1,2,3]
    for i: uint256 in a:
        self.foo(a)
    """
    vyper_module = parse_to_ast(code)
    analyze_module(vyper_module, dummy_input_bundle)


def test_modify_iterator(dummy_input_bundle):
    code = """

a: uint256[3]

@internal
def bar():
    for i: uint256 in self.a:
        self.a[0] = 1
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(ImmutableViolation):
        analyze_module(vyper_module, dummy_input_bundle)


def test_bad_keywords(dummy_input_bundle):
    code = """

@internal
def bar(n: uint256):
    x: uint256 = 0
    for i: uint256 in range(n, boundddd=10):
        x += i
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(ArgumentException):
        analyze_module(vyper_module, dummy_input_bundle)


def test_bad_bound(dummy_input_bundle):
    code = """

@internal
def bar(n: uint256):
    x: uint256 = 0
    for i: uint256 in range(n, bound=n):
        x += i
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(StructureException):
        analyze_module(vyper_module, dummy_input_bundle)


def test_modify_iterator_function_call(dummy_input_bundle):
    code = """

a: uint256[3]

@internal
def foo():
    self.a[0] = 1

@internal
def bar():
    for i: uint256 in self.a:
        self.foo()
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(ImmutableViolation):
        analyze_module(vyper_module, dummy_input_bundle)


def test_modify_iterator_recursive_function_call(dummy_input_bundle):
    code = """

a: uint256[3]

@internal
def foo():
    self.a[0] = 1

@internal
def bar():
    self.foo()

@internal
def baz():
    for i: uint256 in self.a:
        self.bar()
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(ImmutableViolation):
        analyze_module(vyper_module, dummy_input_bundle)


def test_modify_iterator_recursive_function_call_topsort(dummy_input_bundle):
    # test the analysis works no matter the order of functions
    code = """
a: uint256[3]

@internal
def baz():
    for i: uint256 in self.a:
        self.bar()

@internal
def bar():
    self.foo()

@internal
def foo():
    self.a[0] = 1
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(ImmutableViolation) as e:
        analyze_module(vyper_module, dummy_input_bundle)

    assert e.value._message == "Cannot modify loop variable `a`"


def test_modify_iterator_through_struct(dummy_input_bundle):
    # GH issue 3429
    code = """
struct A:
    iter: DynArray[uint256, 5]

a: A

@external
def foo():
    self.a.iter = [1, 2, 3]
    for i: uint256 in self.a.iter:
        self.a = A(iter=[1, 2, 3, 4])
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(ImmutableViolation) as e:
        analyze_module(vyper_module, dummy_input_bundle)

    assert e.value._message == "Cannot modify loop variable `a`"


def test_modify_iterator_complex_expr(dummy_input_bundle):
    # GH issue 3429
    # avoid false positive!
    code = """
a: DynArray[uint256, 5]
b: uint256[10]

@external
def foo():
    self.a = [1, 2, 3]
    for i: uint256 in self.a:
        self.b[self.a[1]] = i
    """
    vyper_module = parse_to_ast(code)
    analyze_module(vyper_module, dummy_input_bundle)


def test_modify_iterator_siblings(dummy_input_bundle):
    # test we can modify siblings in an access tree
    code = """
struct Foo:
    a: uint256[2]
    b: uint256

f: Foo

@external
def foo():
    for i: uint256 in self.f.a:
        self.f.b += i
    """
    vyper_module = parse_to_ast(code)
    analyze_module(vyper_module, dummy_input_bundle)


def test_modify_subscript_barrier(dummy_input_bundle):
    # test that Subscript nodes are a barrier for analysis
    code = """
struct Foo:
    x: uint256[2]
    y: uint256

struct Bar:
    f: Foo[2]

b: Bar

@external
def foo():
    for i: uint256 in self.b.f[1].x:
        self.b.f[0].y += i
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(ImmutableViolation) as e:
        analyze_module(vyper_module, dummy_input_bundle)

    assert e.value._message == "Cannot modify loop variable `b`"


iterator_inference_codes = [
    """
@external
def main():
    for j: uint256 in range(3):
        x: uint256 = j
        y: uint16 = j
    """,  # GH issue 3212
    """
@external
def foo():
    for i: uint256 in [1]:
        a: uint256 = i
        b: uint16 = i
    """,  # GH issue 3374
    """
@external
def foo():
    for i: uint256 in [1]:
        for j: uint256 in [1]:
            a: uint256 = i
        b: uint16 = i
    """,  # GH issue 3374
    """
@external
def foo():
    for i: uint256 in [1,2,3]:
        for j: uint256 in [1,2,3]:
            b: uint256 = j + i
        c: uint16 = i
    """,  # GH issue 3374
]


@pytest.mark.parametrize("code", iterator_inference_codes)
def test_iterator_type_inference_checker(code, dummy_input_bundle):
    vyper_module = parse_to_ast(code)
    with pytest.raises(TypeMismatch):
        analyze_module(vyper_module, dummy_input_bundle)



import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from tests.utils import parse_and_fold
from vyper.exceptions import UnfoldableNode


# TODO expand to all signed types
@pytest.mark.fuzzing
@settings(max_examples=50)
@given(left=st.integers(), right=st.integers())
@pytest.mark.parametrize("op", ["==", "!=", "<", "<=", ">=", ">"])
def test_compare_eq_signed(get_contract, op, left, right):
    source = f"""
@external
def foo(a: int128, b: int128) -> bool:
    return a {op} b
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"{left} {op} {right}")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(left, right) == new_node.value


# TODO expand to all unsigned types
@pytest.mark.fuzzing
@settings(max_examples=50)
@given(left=st.integers(min_value=0), right=st.integers(min_value=0))
@pytest.mark.parametrize("op", ["==", "!=", "<", "<=", ">=", ">"])
def test_compare_eq_unsigned(get_contract, op, left, right):
    source = f"""
@external
def foo(a: uint128, b: uint128) -> bool:
    return a {op} b
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"{left} {op} {right}")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(left, right) == new_node.value


@pytest.mark.fuzzing
@settings(max_examples=20)
@given(left=st.integers(), right=st.lists(st.integers(), min_size=1, max_size=16))
def test_compare_in(left, right, get_contract):
    source = f"""
@external
def foo(a: int128, b: int128[{len(right)}]) -> bool:
    c: int128[{len(right)}] = b
    return a in c

@external
def bar(a: int128) -> bool:
    # note: codegen unrolls to `a == right[0] or a == right[1] ...`
    return a in {right}
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"{left} in {right}")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    # check runtime == fully folded
    assert contract.foo(left, right) == new_node.value
    # check unrolled runtime == fully folded
    assert contract.bar(left) == new_node.value
    # check folding matches python semantics
    assert (left in right) == new_node.value


@pytest.mark.fuzzing
@settings(max_examples=20)
@given(left=st.integers(), right=st.lists(st.integers(), min_size=1, max_size=16))
def test_compare_not_in(left, right, get_contract):
    source = f"""
@external
def foo(a: int128, b: int128[{len(right)}]) -> bool:
    c: int128[{len(right)}] = b
    return a not in c

@external
def bar(a: int128) -> bool:
    # note: codegen unrolls to `a != right[0] and a != right[1] ...`
    return a not in {right}
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"{left} not in {right}")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    # check runtime == fully folded
    assert contract.foo(left, right) == new_node.value
    # check unrolled runtime == fully folded
    assert contract.bar(left) == new_node.value
    # check folding matches python semantics
    assert (left not in right) == new_node.value


@pytest.mark.parametrize("op", ["==", "!=", "<", "<=", ">=", ">"])
def test_compare_type_mismatch(op):
    vyper_ast = parse_and_fold(f"1 {op} 1.0")
    old_node = vyper_ast.body[0].value
    with pytest.raises(UnfoldableNode):
        old_node.get_folded_value()



import math

import pytest

from vyper.compiler import compile_code
from vyper.exceptions import InvalidType, OverflowException

VALID_BITS = list(range(8, 257, 8))


@pytest.mark.parametrize("bits", VALID_BITS)
def test_mkstr(get_contract, bits):
    n_digits = math.ceil(bits * math.log(2) / math.log(10))
    code = f"""
@external
def foo(inp: uint{bits}) -> String[{n_digits}]:
    return uint2str(inp)
    """

    c = get_contract(code)
    for i in [1, 2, 2**bits - 1, 0]:
        assert c.foo(i) == str(i), (i, c.foo(i))


# test for buffer overflow
@pytest.mark.parametrize("bits", VALID_BITS)
def test_mkstr_buffer(get_contract, bits):
    n_digits = math.ceil(bits * math.log(2) / math.log(10))
    code = f"""
some_string: String[{n_digits}]
@internal
def _foo(x: uint{bits}):
    self.some_string = uint2str(x)

@external
def foo(x: uint{bits}) -> uint256:
    y: uint256 = 0
    self._foo(x)
    return y
    """
    c = get_contract(code)
    assert c.foo(2**bits - 1) == 0, bits


def test_bignum_throws():
    code = """
@external
def test():
    a: String[78] = uint2str(2**256)
    pass
    """
    with pytest.raises(OverflowException):
        compile_code(code)


def test_int_fails():
    code = """
@external
def test():
    a: String[78] = uint2str(-1)
    pass
    """
    with pytest.raises(InvalidType):
        compile_code(code)



def test_name_registry(env, get_contract, tx_failed):
    a0, a1 = env.accounts[:2]
    with open("examples/name_registry/name_registry.vy") as f:
        code = f.read()
    c = get_contract(code)
    c.register(b"jacques", a0)
    assert c.lookup(b"jacques") == a0
    with tx_failed():
        c.register(b"jacques", a1)



import pytest
from hypothesis import example, given, settings
from hypothesis import strategies as st

from vyper.codegen.arithmetic import calculate_largest_base, calculate_largest_power
from vyper.compiler import compile_code
from vyper.exceptions import InvalidLiteral


def test_compiler_hang():
    code = """
@external
def f0():
    lv0: uint256 = max_value(int128) ** max_value(int128)
    """
    with pytest.raises(InvalidLiteral):
        compile_code(code)


def test_fold_nonliteral(get_contract):
    # test we can fold non-literal constants
    code = """
N: constant(uint256) = 3
N2: public(constant(uint256)) = N**N
    """
    c = get_contract(code)
    assert c.N2() == 3**3


@pytest.mark.fuzzing
@pytest.mark.parametrize("power", range(2, 255))
def test_exp_uint256(get_contract, tx_failed, power):
    code = f"""
@external
def foo(a: uint256) -> uint256:
    return a ** {power}
    """
    _min_base, max_base = calculate_largest_base(power, 256, False)
    assert max_base**power < 2**256
    assert (max_base + 1) ** power >= 2**256

    c = get_contract(code)

    c.foo(max_base)
    with tx_failed():
        c.foo(max_base + 1)


@pytest.mark.fuzzing
@pytest.mark.parametrize("power", range(2, 127))
def test_exp_int128(get_contract, tx_failed, power):
    code = f"""
@external
def foo(a: int128) -> int128:
    return a ** {power}
    """
    min_base, max_base = calculate_largest_base(power, 128, True)

    assert -(2**127) <= max_base**power < 2**127
    assert -(2**127) <= min_base**power < 2**127

    assert not -(2**127) <= (max_base + 1) ** power < 2**127
    assert not -(2**127) <= (min_base - 1) ** power < 2**127

    c = get_contract(code)

    c.foo(max_base)
    c.foo(min_base)

    with tx_failed():
        c.foo(max_base + 1)
    with tx_failed():
        c.foo(min_base - 1)


@pytest.mark.fuzzing
@pytest.mark.parametrize("power", range(2, 15))
def test_exp_int16(get_contract, tx_failed, power):
    code = f"""
@external
def foo(a: int16) -> int16:
    return a ** {power}
    """
    min_base, max_base = calculate_largest_base(power, 16, True)

    assert -(2**15) <= max_base**power < 2**15
    assert -(2**15) <= min_base**power < 2**15

    assert not -(2**15) <= (max_base + 1) ** power < 2**15
    assert not -(2**15) <= (min_base - 1) ** power < 2**15

    c = get_contract(code)

    c.foo(max_base)
    c.foo(min_base)

    with tx_failed():
        c.foo(max_base + 1)
    with tx_failed():
        c.foo(min_base - 1)


@pytest.mark.fuzzing
@given(a=st.integers(min_value=2, max_value=2**256 - 1))
# 8 bits
@example(a=2**7)
@example(a=2**7 - 1)
# 16 bits
@example(a=2**15)
@example(a=2**15 - 1)
# 32 bits
@example(a=2**31)
@example(a=2**31 - 1)
# 64 bits
@example(a=2**63)
@example(a=2**63 - 1)
# 128 bits
@example(a=2**127)
@example(a=2**127 - 1)
# 256 bits
@example(a=2**256 - 1)
@settings(max_examples=200)
def test_max_exp(get_contract, tx_failed, a):
    code = f"""
@external
def foo(b: uint256) -> uint256:
    return {a} ** b
    """

    c = get_contract(code)

    max_power = calculate_largest_power(a, 256, False)

    assert a**max_power < 2**256
    assert a ** (max_power + 1) >= 2**256

    c.foo(max_power)
    with tx_failed():
        c.foo(max_power + 1)


@pytest.mark.fuzzing
@given(a=st.integers(min_value=2, max_value=2**127 - 1))
# 8 bits
@example(a=2**7)
@example(a=2**7 - 1)
# 16 bits
@example(a=2**15)
@example(a=2**15 - 1)
# 32 bits
@example(a=2**31)
@example(a=2**31 - 1)
# 64 bits
@example(a=2**63)
@example(a=2**63 - 1)
# 128 bits
@example(a=2**127 - 1)
@settings(max_examples=200)
def test_max_exp_int128(get_contract, tx_failed, a):
    code = f"""
@external
def foo(b: int128) -> int128:
    return {a} ** b
    """

    c = get_contract(code)

    max_power = calculate_largest_power(a, 128, True)

    assert -(2**127) <= a**max_power < 2**127
    assert not -(2**127) <= a ** (max_power + 1) < 2**127

    c.foo(max_power)
    with tx_failed():
        c.foo(max_power + 1)



import pytest

from vyper import compiler
from vyper.exceptions import InvalidOperation, StructureException, SyntaxException, TypeMismatch

fail_list = [
    (
        """
@external
def baa():
    x: Bytes[50] = b""
    y: Bytes[50] = b""
    z: Bytes[50] = x + y
    """,
        InvalidOperation,
    ),
    """
@external
def baa():
    x: Bytes[50] = b""
    y: int128 = 0
    y = x
    """,
    """
@external
def baa():
    x: Bytes[50] = b""
    y: int128 = 0
    x = y
    """,
    """
@external
def baa():
    x: Bytes[50] = b""
    y: Bytes[60] = b""
    x = y
    """,
    """
@external
def foo(x: Bytes[100]) -> Bytes[75]:
    return x
    """,
    """
@external
def foo(x: Bytes[100]) -> int128:
    return x
    """,
    """
@external
def foo(x: int128) -> Bytes[75]:
    return x
    """,
    (
        """
@external
def foo() -> Bytes[10]:
    x: Bytes[10] = '0x1234567890123456789012345678901234567890'
    x = 0x1234567890123456789012345678901234567890
    return x
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo() -> Bytes[10]:
    return "badmintonzz"
    """,
        TypeMismatch,
    ),
    (
        """
@external
def test() -> Bytes[1]:
    a: Bytes[1] = 0b0000001  # needs multiple of 8 bits.
    return a
    """,
        SyntaxException,
    ),
    (
        """
@external
def foo():
    a: Bytes = b"abc"
    """,
        StructureException,
    ),
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_bytes_fail(bad_code):
    if isinstance(bad_code, tuple):
        with pytest.raises(bad_code[1]):
            compiler.compile_code(bad_code[0])
    else:
        with pytest.raises(TypeMismatch):
            compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo(x: Bytes[100]) -> Bytes[100]:
    return x
    """,
    """
@external
def foo(x: Bytes[100]) -> Bytes[150]:
    return x
    """,
    """
@external
def baa():
    x: Bytes[50] = b""
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_bytes_success(good_code):
    assert compiler.compile_code(good_code) is not None



from vyper.exceptions import InvalidLiteral, SyntaxException


def test_no_none_assign(assert_compile_failed, get_contract):
    contracts = [  # noqa: E122
        """
@external
def foo():
    bar: int128 = 0
    bar = None
    """,
        """
@external
def foo():
    bar: uint256 = 0
    bar = None
    """,
        """
@external
def foo():
    bar: bool = False
    bar = None
    """,
        """
@external
def foo():
    bar: decimal = 0.0
    bar = None
    """,
        """
@external
def foo():
    bar: bytes32 = empty(bytes32)
    bar = None
    """,
        """
@external
def foo():
    bar: address = empty(address)
    bar = None
    """,
        """
@external
def foo():
    bar: int128 = None
    """,
        """
@external
def foo():
    bar: uint256 = None
    """,
        """
@external
def foo():
    bar: bool = None
    """,
        """
@external
def foo():
    bar: decimal = None
    """,
        """
@external
def foo():
    bar: bytes32 = None
    """,
        """
@external
def foo():
    bar: address = None
    """,
    ]

    for contract in contracts:
        assert_compile_failed(lambda c=contract: get_contract(c), InvalidLiteral)


def test_no_is_none(assert_compile_failed, get_contract):
    contracts = [  # noqa: E122
        """
@external
def foo():
    bar: int128 = 0
    assert bar is None
    """,
        """
@external
def foo():
    bar: uint256 = 0
    assert bar is None
    """,
        """
@external
def foo():
    bar: bool = False
    assert bar is None
    """,
        """
@external
def foo():
    bar: decimal = 0.0
    assert bar is None
    """,
        """
@external
def foo():
    bar: bytes32 = empty(bytes32)
    assert bar is None
    """,
        """
@external
def foo():
    bar: address = empty(address)
    assert bar is None
    """,
    ]

    for contract in contracts:
        assert_compile_failed(lambda c=contract: get_contract(c), SyntaxException)


def test_no_eq_none(assert_compile_failed, get_contract):
    contracts = [  # noqa: E122
        """
@external
def foo():
    bar: int128 = 0
    assert bar == None
    """,
        """
@external
def foo():
    bar: uint256 = 0
    assert bar == None
    """,
        """
@external
def foo():
    bar: bool = False
    assert bar == None
    """,
        """
@external
def foo():
    bar: decimal = 0.0
    assert bar == None
    """,
        """
@external
def foo():
    bar: bytes32 = empty(bytes32)
    assert bar == None
    """,
        """
@external
def foo():
    bar: address = empty(address)
    assert bar == None
    """,
    ]

    for contract in contracts:
        assert_compile_failed(lambda c=contract: get_contract(c), InvalidLiteral)


def test_struct_none(assert_compile_failed, get_contract):
    contracts = [  # noqa: E122
        """
struct Mom:
    a: uint256
    b: int128

@external
def foo():
    mom: Mom = Mom(a=None, b=0)
    """,
        """
struct Mom:
    a: uint256
    b: int128

@external
def foo():
    mom: Mom = Mom(a=0, b=None)
    """,
        """
struct Mom:
    a: uint256
    b: int128

@external
def foo():
    mom: Mom = Mom(a=None, b=None)
    """,
    ]

    for contract in contracts:
        assert_compile_failed(lambda c=contract: get_contract(c), InvalidLiteral)
